/*
 *  open_vznet.c
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

/*
 * Virtual Networking device used to change VE ownership on packets
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/seq_file.h>

#include <linux/inet.h>
#include <net/ip.h>
#include <linux/skbuff.h>
#include <linux/venet.h>

void veip_stop(struct ve_struct *ve)
{
	struct list_head *p, *tmp;

	write_lock_irq(&veip_hash_lock);
	if (ve->veip == NULL)
		goto unlock;
	list_for_each_safe(p, tmp, &ve->veip->ip_lh) {
		struct ip_entry_struct *ptr;
		ptr = list_entry(p, struct ip_entry_struct, ve_list);
		ptr->active_env = NULL;
		list_del(&ptr->ve_list);
		list_del(&ptr->ip_hash);
		kfree(ptr);
	}
	veip_put(ve->veip);
	ve->veip = NULL;
unlock:
	write_unlock_irq(&veip_hash_lock);
}

int veip_start(struct ve_struct *ve)
{
	int err;

	err = 0;
	write_lock_irq(&veip_hash_lock);
	ve->veip = veip_findcreate(ve->veid);
	if (ve->veip == NULL)
		err = -ENOMEM;
	write_unlock_irq(&veip_hash_lock);
	return err;
}

int veip_entry_add(struct ve_struct *ve, struct sockaddr *addr)
{
	struct ip_entry_struct *entry, *found;
	int err;

	entry = kmalloc(sizeof(struct ip_entry_struct), GFP_KERNEL);
	if (entry == NULL)
		return -ENOMEM;

	memset(entry, 0, sizeof(struct ip_entry_struct));
	entry->family = addr->sa_family;
	if (addr->sa_family == AF_INET) {
		entry->key[3] = ((struct sockaddr_in*)addr)->sin_addr.s_addr;
	} else if (addr->sa_family == AF_INET6) {
		memcpy(entry->key, &((struct sockaddr_in6*)addr)->sin6_addr, 16);
	} else {
		kfree(entry);
		return -EAFNOSUPPORT;
	}

	write_lock_irq(&veip_hash_lock);
	err = -EADDRINUSE;
	found = venet_entry_lookup(entry->key, entry->family);
	if (found != NULL)
		goto out_unlock;
	else {
		ip_entry_hash(entry, ve->veip);
		found = entry;
		entry = NULL;
	}
	err = 0;
	found->active_env = ve;
out_unlock:
	write_unlock_irq(&veip_hash_lock);
	if (entry != NULL)
		kfree(entry);
	return err;
}

int veip_entry_del(envid_t veid, struct sockaddr *addr)
{
	struct ip_entry_struct *found;
	u32 key[4];
	int err;

	if (addr->sa_family == AF_INET) {
		memset(key, 0, sizeof(key));
		key[3] = ((struct sockaddr_in*)addr)->sin_addr.s_addr;
	} else if (addr->sa_family == AF_INET6) {
		memcpy(key, &((struct sockaddr_in6*)addr)->sin6_addr, 16);
	} else {
		return -EAFNOSUPPORT;
	}

	err = -EADDRNOTAVAIL;
	write_lock_irq(&veip_hash_lock);
	found = venet_entry_lookup(key, addr->sa_family);
	if (found == NULL)
		goto out;
	if (found->active_env->veid != veid)
		goto out;

	err = 0;
	found->active_env = NULL;

	list_del(&found->ip_hash);
	list_del(&found->ve_list);
	kfree(found);
out:
	write_unlock_irq(&veip_hash_lock);
	return err;
}

static struct ve_struct *venet_find_ve(struct sk_buff *skb, int dir)
{
	struct ip_entry_struct *entry;

	if (skb->protocol == __constant_htons(ETH_P_IP)) {
		entry = ip_entry_lookup(dir ? skb->nh.iph->daddr :
					skb->nh.iph->saddr);
#if defined(CONFIG_IPV6) || defined(CONFIG_IPV6_MODULE)
	} else if (skb->protocol == __constant_htons(ETH_P_IPV6)) {
		entry = venet_entry_lookup(dir ? skb->nh.ipv6h->daddr.s6_addr32 :
					   skb->nh.ipv6h->saddr.s6_addr32, AF_INET6);
#endif
	} else {
		return NULL;
	}
	if (entry == NULL)
		return NULL;

	return entry->active_env;
}

int venet_change_skb_owner(struct sk_buff *skb)
{
	struct ve_struct *ve, *ve_old;

	ve_old = skb->owner_env;

	read_lock(&veip_hash_lock);
	if (!ve_is_super(ve_old)) {
		/* from VE to host */
		ve = venet_find_ve(skb, 0);
		if (ve == NULL)
			goto out_drop;
		if (!ve_accessible_strict(ve, ve_old))
			goto out_source;
		skb->owner_env = get_ve0();
	} else {
		/* from host to VE */
		ve = venet_find_ve(skb, 1);
		if (ve == NULL)
			goto out_drop;
		skb->owner_env = ve;
	}
	read_unlock(&veip_hash_lock);

	return 0;

out_drop:
	read_unlock(&veip_hash_lock);
	return -ESRCH;

out_source:
	read_unlock(&veip_hash_lock);
	if (net_ratelimit() && skb->protocol == __constant_htons(ETH_P_IP)) {
		printk(KERN_WARNING "Dropped packet, source wrong "
		       "veid=%u src-IP=%u.%u.%u.%u "
		       "dst-IP=%u.%u.%u.%u\n",
		       skb->owner_env->veid,
		       NIPQUAD(skb->nh.iph->saddr),
		       NIPQUAD(skb->nh.iph->daddr));
	}
	return -EACCES;
}

#ifdef CONFIG_PROC_FS
int veip_seq_show(struct seq_file *m, void *v)
{
	struct list_head *p;
	struct ip_entry_struct *entry;
	char s[40];

	p = (struct list_head *)v;
	if (p == ip_entry_hash_table) {
		seq_puts(m, "Version: 2.5\n");
		return 0;
	}
	entry = list_entry(p, struct ip_entry_struct, ip_hash);
	if (entry->family == AF_INET)
		sprintf(s, "%u.%u.%u.%u", NIPQUAD(entry->key[3]));
	else
		sprintf(s, "%x:%x:%x:%x:%x:%x:%x:%x",
			ntohl(entry->key[0])>>16,
			ntohl(entry->key[0])&0xFFFF,
			ntohl(entry->key[1])>>16,
			ntohl(entry->key[1])&0xFFFF,
			ntohl(entry->key[2])>>16,
			ntohl(entry->key[2])&0xFFFF,
			ntohl(entry->key[3])>>16,
			ntohl(entry->key[3])&0xFFFF);
	seq_printf(m, "%39s %10u\n", s, 0);
	return 0;
}
#endif

MODULE_AUTHOR("SWsoft <info@sw-soft.com>");
MODULE_DESCRIPTION("Virtuozzo Virtual Network Device");
MODULE_LICENSE("GPL v2");
