/*
 *  veth.c
 *
 *  Copyright (C) 2006  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

/*
 * Virtual ethernet device used to change VE ownership on packets
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/string.h>
#include <linux/socket.h>
#include <linux/errno.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/tcp.h>
#include <linux/proc_fs.h>
#include <linux/seq_file.h>

#include <asm/system.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/unistd.h>

#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <net/ip.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <linux/if_ether.h>	/* For the statistics structure. */
#include <linux/if_arp.h>	/* For ARPHRD_ETHER */
#include <linux/ve_proto.h>
#include <linux/vzctl.h>
#include <linux/vzctl_veth.h>

#include <linux/list.h>
#include <linux/spinlock.h>
#include <linux/vzcalluser.h>

struct veth_struct
{
	struct net_device_stats stats;
	struct net_device	*pair;
	struct list_head	hwaddr_list;
};

struct list_head veth_hwaddr_list;
rwlock_t ve_hwaddr_lock = RW_LOCK_UNLOCKED;
DECLARE_MUTEX(hwaddr_sem);

#define veth_from_netdev(dev) \
	((struct veth_struct *)(netdev_priv(dev)))
#define veth_to_netdev(veth) \
	((struct net_device*)((char*)veth - \
	(unsigned long)netdev_priv(NULL)))

struct net_device * veth_dev_start(char *dev_addr, char *name);

struct veth_struct *hwaddr_entry_lookup(char *name)
{
	struct veth_struct *entry;
	struct list_head *tmp;

	list_for_each(tmp, &veth_hwaddr_list) {
		entry = list_entry(tmp, struct veth_struct, hwaddr_list);
		BUG_ON(entry->pair == NULL);
		if (strncmp(name, entry->pair->name, IFNAMSIZ) == 0)
			return entry;
	}
	return NULL;
}

int veth_entry_add(struct ve_struct *ve, char *dev_addr, char *name,
		char *dev_addr_ve, char *name_ve)
{
	struct net_device *dev_ve;
	struct net_device *dev_ve0;
	struct ve_struct *old_env;
	char dev_name[IFNAMSIZ];
	int err;

	down(&hwaddr_sem);

	if (name[0] == '\0')
		snprintf(dev_name, sizeof(dev_name), "vz%d.%%d", ve->veid);
	else {
		memcpy(dev_name, name, IFNAMSIZ - 1);
		dev_name[IFNAMSIZ - 1] = '\0';
	}
	dev_ve0 = veth_dev_start(dev_addr, dev_name);
	if (IS_ERR(dev_ve0)) {
		err = PTR_ERR(dev_ve0);
		goto err;
	}

	old_env = set_exec_env(ve);
	if (name_ve[0] == '\0')
		sprintf(dev_name, "eth%%d");
	else {
		memcpy(dev_name, name_ve, IFNAMSIZ - 1);
		dev_name[IFNAMSIZ - 1] = '\0';
	}
	dev_ve = veth_dev_start(dev_addr_ve, dev_name);
	if (IS_ERR(dev_ve)) {
		err = PTR_ERR(dev_ve);
		goto err_ve;
	}
	set_exec_env(old_env);
	veth_from_netdev(dev_ve)->pair = dev_ve0;
	veth_from_netdev(dev_ve0)->pair = dev_ve;

	write_lock(&ve_hwaddr_lock);
	list_add(&(veth_from_netdev(dev_ve)->hwaddr_list), &veth_hwaddr_list);
	write_unlock(&ve_hwaddr_lock);

	up(&hwaddr_sem);
	return 0;

err_ve:
	set_exec_env(old_env);
	unregister_netdev(dev_ve0);
err:
	up(&hwaddr_sem);
	return err;
}

int veth_entry_del(struct ve_struct *ve, char *name)
{
	struct veth_struct *found;
	struct ve_struct *old_env;
	struct net_device *dev;
	int err;

	err = -ENODEV;
	down(&hwaddr_sem);
	found = hwaddr_entry_lookup(name);
	if (found == NULL)
		goto out;
	if (veth_to_netdev(found)->owner_env != ve)
		goto out;

	write_lock(&ve_hwaddr_lock);
	list_del(&found->hwaddr_list);
	write_unlock(&ve_hwaddr_lock);
	err = 0;
	dev = found->pair;
	BUG_ON(found->pair == NULL);

	old_env = get_exec_env();
	set_exec_env(ve);
	unregister_netdev(veth_to_netdev(found));
	set_exec_env(old_env);

	unregister_netdev(dev);

out:
	up(&hwaddr_sem);
	return err;
}

/*
 * Device functions
 */

static int veth_open(struct net_device *dev)
{
	return 0;
}

static int veth_close(struct net_device *master)
{
	return 0;
}

static void veth_destructor(struct net_device *dev)
{
	free_netdev(dev);
}

static struct net_device_stats *get_stats(struct net_device *dev)
{
	return &veth_from_netdev(dev)->stats;
}

/*
 * The higher levels take care of making this non-reentrant (it's
 * called with bh's disabled).
 */
static int veth_xmit(struct sk_buff *skb, struct net_device *dev)
{
	struct net_device_stats *stats = get_stats(dev);
	struct net_device *rcv = NULL;
	struct veth_struct *entry;
	int length;

	if (unlikely(get_exec_env()->disable_net))
		goto outf;

	skb_orphan(skb);

	entry = veth_from_netdev(dev);
	rcv = entry->pair;
	if (!rcv)
		/* VE going down */
		goto outf;

	if (unlikely(rcv->owner_env->disable_net))
		goto outf;

	skb->owner_env = rcv->owner_env;

	if (!(rcv->flags & IFF_UP)) {
		/* Target VE does not want to receive packets */
		goto outf;
	}

	skb->dev = rcv;
	skb->pkt_type = PACKET_HOST;
	skb->protocol = eth_type_trans(skb, rcv);

	dst_release(skb->dst);
	skb->dst = NULL;
#ifdef CONFIG_NETFILTER
	nf_conntrack_put(skb->nfct);
	skb->nfct = NULL;
#ifdef CONFIG_NETFILTER_DEBUG
	skb->nf_debug = 0;
#endif
#endif
	length = skb->len;

	netif_rx(skb);

	stats->tx_bytes += length;
	stats->tx_packets++;
	if (rcv) {
		struct net_device_stats *rcv_stats = get_stats(rcv);
		rcv_stats->rx_bytes += length;
		rcv_stats->rx_packets++;
	}

	return 0;

outf:
	kfree_skb(skb);
	stats->tx_dropped++;
	return 0;
}

int veth_init_dev(struct net_device *dev)
{
	dev->hard_start_xmit = veth_xmit;
	dev->get_stats = get_stats;
	dev->open = veth_open;
	dev->stop = veth_close;
	dev->destructor = veth_destructor;

	ether_setup(dev);

	dev->tx_queue_len = 0;
	return 0;
}

static void veth_setup(struct net_device *dev)
{
	dev->init = veth_init_dev;
	/*
	 * No other features, as they are:
	 *  - checksumming is required, and nobody else will done our job
	 */
	dev->features |= NETIF_F_VENET | NETIF_F_VIRTUAL;
}

#ifdef CONFIG_PROC_FS
#define ADDR_FMT "%02x:%02x:%02x:%02x:%02x:%02x"
#define ADDR(x) (x)[0],(x)[1],(x)[2],(x)[3],(x)[4],(x)[5]
static int vehwaddr_seq_show(struct seq_file *m, void *v)
{
	struct list_head *p;
	struct veth_struct *entry;

	p = (struct list_head *)v;
	if (p == &veth_hwaddr_list) {
		seq_puts(m, "Version: 1.0\n");
		return 0;
	}
	entry = list_entry(p, struct veth_struct, hwaddr_list);
	seq_printf(m, ADDR_FMT " %16s ",
			ADDR(entry->pair->dev_addr), entry->pair->name);
	seq_printf(m, ADDR_FMT " %16s %10u\n",
			ADDR(veth_to_netdev(entry)->dev_addr),
			veth_to_netdev(entry)->name,
			VEID(veth_to_netdev(entry)->owner_env));
	return 0;
}

static void *vehwaddr_seq_start(struct seq_file *m, loff_t *pos)
{
	loff_t l;
	struct list_head *p;

	l = *pos;
	read_lock(&ve_hwaddr_lock);
	if (l == 0)
		return &veth_hwaddr_list;
	list_for_each(p, &veth_hwaddr_list) {
		if (--l == 0)
			return p;
	}
	return NULL;
}

static void *vehwaddr_seq_next(struct seq_file *m, void *v, loff_t *pos)
{
	struct list_head *p;

	p = (struct list_head *)v;
	(*pos)++;
	return p->next == &veth_hwaddr_list ? NULL : p->next;
}

static void vehwaddr_seq_stop(struct seq_file *m, void *v)
{
	read_unlock(&ve_hwaddr_lock);
}

static struct seq_operations vehwaddr_seq_op = {
	.start 	= vehwaddr_seq_start,
	.next	= vehwaddr_seq_next,
	.stop	= vehwaddr_seq_stop,
	.show	= vehwaddr_seq_show
};

static int vehwaddr_open(struct inode *inode, struct file *file)
{
	return seq_open(file, &vehwaddr_seq_op);
}

static struct file_operations proc_vehwaddr_operations = {
	.open		= vehwaddr_open,
	.read		= seq_read,
	.llseek		= seq_lseek,
	.release	= seq_release
};
#endif

int real_ve_hwaddr(envid_t veid, int op,
		unsigned char *dev_addr, int addrlen, char *name,
		unsigned char *dev_addr_ve, int addrlen_ve, char *name_ve)
{
	int err;
	struct ve_struct *ve;
	char ve_addr[ETH_ALEN];

	err = -EPERM;
	if (!capable(CAP_NET_ADMIN))
		goto out;

	err = -EINVAL;
	switch (op)
	{
		case VE_ETH_ADD:
			if (addrlen != ETH_ALEN)
				goto out;
			if (addrlen_ve != ETH_ALEN && addrlen_ve != 0)
				goto out;
			/* If ve addr is not set then we use dev_addr[3] & 0x80 for it */
			if (addrlen_ve == 0 && (dev_addr[3] & 0x80))
				goto out;
			if (addrlen_ve == 0) {
				memcpy(ve_addr, dev_addr, ETH_ALEN);
				ve_addr[3] |= 0x80;
			} else {
				memcpy(ve_addr, dev_addr_ve, ETH_ALEN);
			}

			ve = get_ve_by_id(veid);
			err = -ESRCH;
			if (!ve)
				goto out;

			down_read(&ve->op_sem);
			if (ve->is_running)
				err = veth_entry_add(ve, dev_addr, name,
						ve_addr, name_ve);
			up_read(&ve->op_sem);
			put_ve(ve);
			break;

		case VE_ETH_DEL:
			if (name[0] == '\0')
				goto out;
			ve = get_ve_by_id(veid);
			err = -ESRCH;
			if (!ve)
				goto out;

			down_read(&ve->op_sem);
			if (ve->is_running)
				err = veth_entry_del(ve, name);
			up_read(&ve->op_sem);
			put_ve(ve);
			break;
	}

out:
	return err;
}

int veth_ioctl(struct inode *ino, struct file *file, unsigned int cmd,
		unsigned long arg)
{
	int err;

	err = -ENOTTY;
	switch(cmd) {
	    case VETHCTL_VE_HWADDR: {
			struct vzctl_ve_hwaddr s;
			err = -EFAULT;
			if (copy_from_user(&s, (void *)arg, sizeof(s)))
				break;
			err = real_ve_hwaddr(s.veid, s.op,
					s.dev_addr, s.addrlen, s.dev_name,
					s.dev_addr_ve, s.addrlen_ve, s.dev_name_ve);
		}
		break;
	}
	return err;
}

static struct vzioctlinfo vethcalls = {
	.type	= VETHCTLTYPE,
	.func	= veth_ioctl,
	.owner	= THIS_MODULE,
};

struct net_device * veth_dev_start(char *dev_addr, char *name)
{
	struct net_device *dev;
	int err;

	dev = alloc_netdev(sizeof(struct veth_struct), name, veth_setup);
	if (!dev)
		return ERR_PTR(-ENOMEM);
	if (strchr(dev->name, '%')) {
		err = dev_alloc_name(dev, dev->name);
		if (err < 0)
			goto err;
	}
	if ((err = register_netdev(dev)) != 0)
		goto err;

	memcpy(dev->dev_addr, dev_addr, ETH_ALEN);
	dev->addr_len = ETH_ALEN;

	return dev;
err:
	free_netdev(dev);
	printk(KERN_ERR "%s initialization error err=%d\n", name, err);
	return ERR_PTR(err);
}

static int veth_stop(unsigned int hooknum, void *data)
{
	struct ve_struct *old_env;
	struct ve_struct *env;
	struct list_head *tmp, *n;

	env = (struct ve_struct *)data;
	down(&hwaddr_sem);
	list_for_each_safe(tmp, n, &veth_hwaddr_list) {
		struct veth_struct *entry;
		struct net_device *dev;
		entry = list_entry(tmp, struct veth_struct, hwaddr_list);
		if (VEID(env) != VEID(veth_to_netdev(entry)->owner_env))
			continue;

		write_lock(&ve_hwaddr_lock);
		list_del(&entry->hwaddr_list);
		write_unlock(&ve_hwaddr_lock);

		dev = entry->pair;
		BUG_ON(entry->pair == NULL);
		old_env = set_exec_env(env);
		unregister_netdev(veth_to_netdev(entry));
		set_exec_env(old_env);

		old_env = set_exec_env(get_ve0());
		unregister_netdev(dev);
		set_exec_env(old_env);
	}
	up(&hwaddr_sem);
	return 0;
}

#define VE_HOOK_PRI_NET		0

static struct ve_hook veth_ve_hook_fini = {
	.hook		= veth_stop,
	.hooknum	= VE_HOOK_FINI,
	.priority	= VE_HOOK_PRI_NET,
	.owner		= THIS_MODULE,
};

__init int veth_init(void)
{
#ifdef CONFIG_PROC_FS
	struct proc_dir_entry *de;
#endif

	INIT_LIST_HEAD(&veth_hwaddr_list);

#ifdef CONFIG_PROC_FS
	de = create_proc_glob_entry("vz/veth",
			S_IFREG|S_IRUSR, NULL);
	if (de)
		de->proc_fops = &proc_vehwaddr_operations;
	else
		printk(KERN_WARNING "veth: can't make vehwaddr proc entry\n");

#endif

	ve_hook_register(&veth_ve_hook_fini);
	vzioctl_register(&vethcalls);
	return 0;
}

__exit void veth_exit(void)
{
	struct veth_struct *entry;
	struct list_head *tmp, *n;
	struct ve_struct *ve;
	struct ve_struct *old_env;

	vzioctl_unregister(&vethcalls);
	ve_hook_unregister(&veth_ve_hook_fini);
#ifdef CONFIG_PROC_FS
	remove_proc_entry("vz/veth", NULL);
#endif

	down(&hwaddr_sem);
	list_for_each_safe(tmp, n, &veth_hwaddr_list) {
		struct net_device *dev;
		entry = list_entry(tmp, struct veth_struct, hwaddr_list);
		ve = get_ve(veth_to_netdev(entry)->owner_env);

		write_lock(&ve_hwaddr_lock);
		list_del(&entry->hwaddr_list);
		write_unlock(&ve_hwaddr_lock);

		dev = entry->pair;
		BUG_ON(entry->pair == NULL);
		old_env = set_exec_env(ve);
		unregister_netdev(veth_to_netdev(entry));
		set_exec_env(old_env);

		unregister_netdev(dev);

		put_ve(ve);
	}
	up(&hwaddr_sem);
}

module_init(veth_init);
module_exit(veth_exit);

MODULE_AUTHOR("Andrey Mirkin <amirkin@sw.ru>");
MODULE_DESCRIPTION("Virtuozzo Virtual Ethernet Device");
MODULE_LICENSE("GPL v2");

