/*
 *
 *  kernel/cpt/cpt_process.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/compat.h>
#include <linux/cpt_image.h>

#include "cpt_obj.h"
#include "cpt_context.h"
#include "cpt_ubc.h"
#include "cpt_process.h"
#include "cpt_kernel.h"

#ifdef CONFIG_X86_32
#undef task_pt_regs
#define task_pt_regs(t) ((struct pt_regs *)((t)->thread.esp0) - 1)
#endif

static u32 encode_segment(u32 segreg)
{
	segreg &= 0xFFFF;

	if (segreg == 0)
		return CPT_SEG_ZERO;
	if ((segreg & 3) != 3) {
		wprintk("Invalid RPL of a segment reg %x\n", segreg);
		return CPT_SEG_ZERO;
	}

	/* LDT descriptor, it is just an index to LDT array */
	if (segreg & 4)
		return CPT_SEG_LDT + (segreg >> 3);

	/* TLS descriptor. */
	if ((segreg >> 3) >= GDT_ENTRY_TLS_MIN &&
	    (segreg >> 3) <= GDT_ENTRY_TLS_MAX)
		return CPT_SEG_TLS1 + ((segreg>>3) - GDT_ENTRY_TLS_MIN);

	/* One of standard desriptors */
#ifdef CONFIG_X86_64
	if (segreg == __USER32_DS)
		return CPT_SEG_USER32_DS;
	if (segreg == __USER32_CS)
		return CPT_SEG_USER32_CS;
	if (segreg == __USER_DS)
		return CPT_SEG_USER64_DS;
	if (segreg == __USER_CS)
		return CPT_SEG_USER64_CS;
#else
	if (segreg == __USER_DS)
		return CPT_SEG_USER32_DS;
	if (segreg == __USER_CS)
		return CPT_SEG_USER32_CS;
#endif
	wprintk("Invalid segment reg %x\n", segreg);
	return CPT_SEG_ZERO;
}

#ifdef CONFIG_X86_64
static void xlate_ptregs_64_to_32(struct cpt_x86_regs *d, struct pt_regs *s, task_t *tsk)
{
	d->cpt_ebp = s->rbp;
	d->cpt_ebx = s->rbx;
	d->cpt_eax = s->rax;
	d->cpt_ecx = s->rcx;
	d->cpt_edx = s->rdx;
	d->cpt_esi = s->rsi;
	d->cpt_edi = s->rdi;
	d->cpt_orig_eax = s->orig_rax;
	d->cpt_eip = s->rip;
	d->cpt_xcs = encode_segment(s->cs);
	d->cpt_eflags = s->eflags;
	d->cpt_esp = s->rsp;
	d->cpt_xss = encode_segment(s->ss);
	d->cpt_xds = encode_segment(tsk->thread.ds);
	d->cpt_xes = encode_segment(tsk->thread.es);
}

static int dump_registers(task_t *tsk, struct cpt_context *ctx)
{
	cpt_open_object(NULL, ctx);

	if (tsk->thread_info->flags&_TIF_IA32) {
		struct cpt_x86_regs ri;
		ri.cpt_next = sizeof(ri);
		ri.cpt_object = CPT_OBJ_X86_REGS;
		ri.cpt_hdrlen = sizeof(ri);
		ri.cpt_content = CPT_CONTENT_VOID;

		ri.cpt_debugreg[0] = tsk->thread.debugreg0;
		ri.cpt_debugreg[1] = tsk->thread.debugreg1;
		ri.cpt_debugreg[2] = tsk->thread.debugreg2;
		ri.cpt_debugreg[3] = tsk->thread.debugreg3;
		ri.cpt_debugreg[4] = 0;
		ri.cpt_debugreg[5] = 0;
		ri.cpt_debugreg[6] = tsk->thread.debugreg6;
		ri.cpt_debugreg[7] = tsk->thread.debugreg7;
		ri.cpt_fs = encode_segment(tsk->thread.fsindex);
		ri.cpt_gs = encode_segment(tsk->thread.gsindex);

		xlate_ptregs_64_to_32(&ri, task_pt_regs(tsk), tsk);

		ctx->write(&ri, sizeof(ri), ctx);
	} else {
		struct cpt_x86_64_regs ri;
		ri.cpt_next = sizeof(ri);
		ri.cpt_object = CPT_OBJ_X86_64_REGS;
		ri.cpt_hdrlen = sizeof(ri);
		ri.cpt_content = CPT_CONTENT_VOID;

		ri.cpt_fsbase = tsk->thread.fs;
		ri.cpt_gsbase = tsk->thread.gs;
		ri.cpt_fsindex = encode_segment(tsk->thread.fsindex);
		ri.cpt_gsindex = encode_segment(tsk->thread.gsindex);
		ri.cpt_ds = encode_segment(tsk->thread.ds);
		ri.cpt_es = encode_segment(tsk->thread.es);
		ri.cpt_debugreg[0] = tsk->thread.debugreg0;
		ri.cpt_debugreg[1] = tsk->thread.debugreg1;
		ri.cpt_debugreg[2] = tsk->thread.debugreg2;
		ri.cpt_debugreg[3] = tsk->thread.debugreg3;
		ri.cpt_debugreg[4] = 0;
		ri.cpt_debugreg[5] = 0;
		ri.cpt_debugreg[6] = tsk->thread.debugreg6;
		ri.cpt_debugreg[7] = tsk->thread.debugreg7;

		memcpy(&ri.cpt_r15, task_pt_regs(tsk), sizeof(struct pt_regs));

		ri.cpt_cs = encode_segment(task_pt_regs(tsk)->cs);
		ri.cpt_ss = encode_segment(task_pt_regs(tsk)->ss);

		ctx->write(&ri, sizeof(ri), ctx);

#if 0
		if (ri.cpt_rip >= VSYSCALL_START && ri.cpt_rip < VSYSCALL_END) {
			eprintk_ctx(CPT_FID "cannot be checkpointied while vsyscall, try later\n", CPT_TID(tsk));
			return -EAGAIN;
		}
#endif
	}
	cpt_close_object(ctx);

	return 0;
}

#else

static int dump_registers(task_t *tsk, struct cpt_context *ctx)
{
	struct cpt_x86_regs ri;

	cpt_open_object(NULL, ctx);

	ri.cpt_next = sizeof(ri);
	ri.cpt_object = CPT_OBJ_X86_REGS;
	ri.cpt_hdrlen = sizeof(ri);
	ri.cpt_content = CPT_CONTENT_VOID;

	ri.cpt_debugreg[0] = tsk->thread.debugreg[0];
	ri.cpt_debugreg[1] = tsk->thread.debugreg[1];
	ri.cpt_debugreg[2] = tsk->thread.debugreg[2];
	ri.cpt_debugreg[3] = tsk->thread.debugreg[3];
	ri.cpt_debugreg[4] = tsk->thread.debugreg[4];
	ri.cpt_debugreg[5] = tsk->thread.debugreg[5];
	ri.cpt_debugreg[6] = tsk->thread.debugreg[6];
	ri.cpt_debugreg[7] = tsk->thread.debugreg[7];
	ri.cpt_fs = encode_segment(tsk->thread.fs);
	ri.cpt_gs = encode_segment(tsk->thread.gs);

	memcpy(&ri.cpt_ebx, task_pt_regs(tsk), sizeof(struct pt_regs));

	ri.cpt_xcs = encode_segment(task_pt_regs(tsk)->xcs);
	ri.cpt_xss = encode_segment(task_pt_regs(tsk)->xss);
	ri.cpt_xds = encode_segment(task_pt_regs(tsk)->xds);
        ri.cpt_xes = encode_segment(task_pt_regs(tsk)->xes);

	ctx->write(&ri, sizeof(ri), ctx);
	cpt_close_object(ctx);

	return 0;
}
#endif

static int dump_kstack(task_t *tsk, struct cpt_context *ctx)
{
	struct cpt_obj_bits hdr;
	unsigned long size;
	void *start;

	cpt_open_object(NULL, ctx);

#ifdef CONFIG_X86_64
	size = tsk->thread.rsp0 - tsk->thread.rsp;
	start = (void*)tsk->thread.rsp;
#else
	size = tsk->thread.esp0 - tsk->thread.esp;
	start = (void*)tsk->thread.esp;
#endif

	hdr.cpt_next = sizeof(hdr) + CPT_ALIGN(size);
	hdr.cpt_object = CPT_OBJ_BITS;
	hdr.cpt_hdrlen = sizeof(hdr);
	hdr.cpt_content = CPT_CONTENT_STACK;
	hdr.cpt_size = size;

	ctx->write(&hdr, sizeof(hdr), ctx);
	ctx->write(start, size, ctx);
	ctx->align(ctx);
	cpt_close_object(ctx);
	return 0;
}

/* Formats of i387_fxsave_struct are the same for x86_64
 * and i386. Plain luck. */

static int dump_fpustate(task_t *tsk, struct cpt_context *ctx)
{
	struct cpt_obj_bits hdr;
	unsigned long size;
	int type;

	cpt_open_object(NULL, ctx);

	type = CPT_CONTENT_X86_FPUSTATE;
	size = sizeof(struct i387_fxsave_struct);
#ifndef CONFIG_X86_64
	if (!cpu_has_fxsr) {
		size = sizeof(struct i387_fsave_struct);
		type = CPT_CONTENT_X86_FPUSTATE_OLD;
	}
#endif

	hdr.cpt_next = sizeof(hdr) + CPT_ALIGN(size);
	hdr.cpt_object = CPT_OBJ_BITS;
	hdr.cpt_hdrlen = sizeof(hdr);
	hdr.cpt_content = type;
	hdr.cpt_size = size;

	ctx->write(&hdr, sizeof(hdr), ctx);
	ctx->write(&tsk->thread.i387, size, ctx);
	ctx->align(ctx);
	cpt_close_object(ctx);
	return 0;
}

static int encode_siginfo(struct cpt_siginfo_image *si, siginfo_t *info)
{
	si->cpt_signo = info->si_signo;
	si->cpt_errno = info->si_errno;
	si->cpt_code = info->si_code;

	switch(si->cpt_code & __SI_MASK) {
	case __SI_TIMER:
		si->cpt_pid = info->si_tid;
		si->cpt_uid = info->si_overrun;
		si->cpt_sigval = cpt_ptr_export(info->_sifields._timer._sigval.sival_ptr);
		si->cpt_utime = info->si_sys_private;
		break;
	case __SI_POLL:
		si->cpt_pid = info->si_band;
		si->cpt_uid = info->si_fd;
		break;
	case __SI_FAULT:
		si->cpt_sigval = cpt_ptr_export(info->si_addr);
#ifdef __ARCH_SI_TRAPNO
		si->cpt_pid = info->si_trapno;
#endif
		break;
	case __SI_CHLD:
		si->cpt_pid = is_virtual_pid(info->si_pid) ? info->si_pid : pid_type_to_vpid(PIDTYPE_PID, info->si_pid);
		si->cpt_uid = info->si_uid;
		si->cpt_sigval = info->si_status;
		si->cpt_stime = info->si_stime;
		si->cpt_utime = info->si_utime;
		break;
	case __SI_KILL:
	case __SI_RT:
	case __SI_MESGQ:
	default:
		si->cpt_pid = is_virtual_pid(info->si_pid) ? info->si_pid : pid_type_to_vpid(PIDTYPE_TGID, info->si_pid);
		si->cpt_uid = info->si_uid;
		si->cpt_sigval = cpt_ptr_export(info->si_ptr);
		break;
	}
	return 0;
}

static int dump_sigqueue(struct sigpending *list, struct cpt_context *ctx)
{
	struct sigqueue *q;
	loff_t saved_obj;

	if (list_empty(&list->list))
		return 0;

	cpt_push_object(&saved_obj, ctx);
	list_for_each_entry(q, &list->list, list) {
		struct cpt_siginfo_image si;

		si.cpt_next = sizeof(si);
		si.cpt_object = CPT_OBJ_SIGINFO;
		si.cpt_hdrlen = sizeof(si);
		si.cpt_content = CPT_CONTENT_VOID;

		si.cpt_qflags = q->flags;
		si.cpt_user = q->user->uid;

		if (encode_siginfo(&si, &q->info))
			return -EINVAL;

		ctx->write(&si, sizeof(si), ctx);
	}
	cpt_pop_object(&saved_obj, ctx);
	return 0;
}



static int dump_one_signal_struct(cpt_object_t *obj, struct cpt_context *ctx)
{
	struct signal_struct *sig = obj->o_obj;
	struct cpt_signal_image *v = cpt_get_buf(ctx);
	task_t *tsk;
	int i;

	cpt_open_object(obj, ctx);

	v->cpt_next = CPT_NULL;
	v->cpt_object = CPT_OBJ_SIGNAL_STRUCT;
	v->cpt_hdrlen = sizeof(*v);
	v->cpt_content = CPT_CONTENT_ARRAY;

	if (sig->pgrp <= 0) {
		eprintk_ctx("bad pgid\n");
		cpt_release_buf(ctx);
		return -EINVAL;
	}
	v->cpt_pgrp_type = CPT_PGRP_NORMAL;
	read_lock(&tasklist_lock);
	tsk = find_task_by_pid_type_ve(PIDTYPE_PID, sig->pgrp);
	if (tsk == NULL)
		v->cpt_pgrp_type = CPT_PGRP_ORPHAN;
	read_unlock(&tasklist_lock);
	v->cpt_pgrp = pid_type_to_vpid(PIDTYPE_PGID, sig->pgrp);

	v->cpt_old_pgrp = 0;
	if (sig->tty_old_pgrp < 0) {
		eprintk_ctx("bad tty_old_pgrp\n");
		cpt_release_buf(ctx);
		return -EINVAL;
	}
	if (sig->tty_old_pgrp > 0) {
		v->cpt_old_pgrp_type = CPT_PGRP_NORMAL;
		read_lock(&tasklist_lock);
		tsk = find_task_by_pid_type_ve(PIDTYPE_PID, sig->tty_old_pgrp);
		if (tsk == NULL) {
			v->cpt_old_pgrp_type = CPT_PGRP_ORPHAN;
			tsk = find_task_by_pid_type_ve(PIDTYPE_PGID, sig->tty_old_pgrp);
		}
		read_unlock(&tasklist_lock);
		if (tsk == NULL) {
			eprintk_ctx("tty_old_pgrp does not exist anymore\n");
			cpt_release_buf(ctx);
			return -EINVAL;
		}
		v->cpt_old_pgrp = _pid_type_to_vpid(PIDTYPE_PGID, sig->tty_old_pgrp);
		if ((int)v->cpt_old_pgrp < 0) {
			dprintk_ctx("stray tty_old_pgrp %d\n", sig->tty_old_pgrp);
			v->cpt_old_pgrp = -1;
			v->cpt_old_pgrp_type = CPT_PGRP_STRAY;
		}
	}

	if (sig->session <= 0) {
		eprintk_ctx("bad session\n");
		cpt_release_buf(ctx);
		return -EINVAL;
	}
	v->cpt_session_type = CPT_PGRP_NORMAL;
	read_lock(&tasklist_lock);
	tsk = find_task_by_pid_type_ve(PIDTYPE_PID, sig->session);
	if (tsk == NULL)
		v->cpt_session_type = CPT_PGRP_ORPHAN;
	read_unlock(&tasklist_lock);
	v->cpt_session = pid_type_to_vpid(PIDTYPE_SID, sig->session);

	v->cpt_leader = sig->leader;
	v->cpt_ctty = CPT_NULL;
	if (sig->tty) {
		cpt_object_t *cobj = lookup_cpt_object(CPT_OBJ_TTY, sig->tty, ctx);
		if (cobj)
			v->cpt_ctty = cobj->o_pos;
		else {
			eprintk_ctx("controlling tty is not found\n");
			cpt_release_buf(ctx);
			return -EINVAL;
		}
	}
	memcpy(&v->cpt_sigpending, &sig->shared_pending.signal, 8);

	v->cpt_curr_target = 0;
	if (sig->curr_target)
		v->cpt_curr_target = virt_pid(sig->curr_target);
	v->cpt_group_exit = ((sig->flags & SIGNAL_GROUP_EXIT) != 0);
	v->cpt_group_exit_code = sig->group_exit_code;
	v->cpt_group_exit_task = 0;
	if (sig->group_exit_task)
		v->cpt_group_exit_task = virt_pid(sig->group_exit_task);
	v->cpt_notify_count = sig->notify_count;
	v->cpt_group_stop_count = sig->group_stop_count;

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,8)
	v->cpt_utime = sig->utime;
	v->cpt_stime = sig->stime;
	v->cpt_cutime = sig->cutime;
	v->cpt_cstime = sig->cstime;
	v->cpt_nvcsw = sig->nvcsw;
	v->cpt_nivcsw = sig->nivcsw;
	v->cpt_cnvcsw = sig->cnvcsw;
	v->cpt_cnivcsw = sig->cnivcsw;
	v->cpt_min_flt = sig->min_flt;
	v->cpt_maj_flt = sig->maj_flt;
	v->cpt_cmin_flt = sig->cmin_flt;
	v->cpt_cmaj_flt = sig->cmaj_flt;

	if (RLIM_NLIMITS > CPT_RLIM_NLIMITS)
		__asm__("undefined\n");

	for (i=0; i<CPT_RLIM_NLIMITS; i++) {
		if (i < RLIM_NLIMITS) {
			v->cpt_rlim_cur[i] = sig->rlim[i].rlim_cur;
			v->cpt_rlim_max[i] = sig->rlim[i].rlim_max;
		} else {
			v->cpt_rlim_cur[i] = CPT_NULL;
			v->cpt_rlim_max[i] = CPT_NULL;
		}
	}
#endif

	ctx->write(v, sizeof(*v), ctx);
	cpt_release_buf(ctx);

	dump_sigqueue(&sig->shared_pending, ctx);

	cpt_close_object(ctx);
	return 0;
}


static int dump_one_process(cpt_object_t *obj, struct cpt_context *ctx)
{
	task_t *tsk = obj->o_obj;
	int last_thread;
	struct cpt_task_image *v = cpt_get_buf(ctx);
	cpt_object_t *tobj;
	cpt_object_t *tg_obj;
	loff_t saved_obj;
	int i;
	int err;
	struct timespec delta;

	cpt_open_object(obj, ctx);

	v->cpt_signal = CPT_NULL;
	tg_obj = lookup_cpt_object(CPT_OBJ_SIGNAL_STRUCT, tsk->signal, ctx);
	if (!tg_obj) BUG();

	v->cpt_next = CPT_NULL;
	v->cpt_object = CPT_OBJ_TASK;
	v->cpt_hdrlen = sizeof(*v);
	v->cpt_content = CPT_CONTENT_ARRAY;

	v->cpt_state = tsk->state;
	if (tsk->state == EXIT_ZOMBIE) {
		eprintk_ctx("invalid zombie state on" CPT_FID "\n", CPT_TID(tsk));
		cpt_release_buf(ctx);
		return -EINVAL;
	} else if (tsk->state == EXIT_DEAD) {
		if (tsk->exit_state != EXIT_DEAD &&
		    tsk->exit_state != EXIT_ZOMBIE) {
			eprintk_ctx("invalid exit_state %ld on" CPT_FID "\n", tsk->exit_state, CPT_TID(tsk));
			cpt_release_buf(ctx);
			return -EINVAL;
		}
	}
	if (tsk->exit_state) {
		v->cpt_state = tsk->exit_state;
		if (tsk->state != EXIT_DEAD) {
			eprintk_ctx("invalid tsk->state %ld/%ld on" CPT_FID "\n",
				tsk->state, tsk->exit_state, CPT_TID(tsk));
			cpt_release_buf(ctx);
			return -EINVAL;
		}
	}
	v->cpt_flags = tsk->flags&~PF_FROZEN;
	v->cpt_ptrace = tsk->ptrace;
	v->cpt_prio = tsk->prio;
	v->cpt_exit_code = tsk->exit_code;
	v->cpt_exit_signal = tsk->exit_signal;
	v->cpt_pdeath_signal = tsk->pdeath_signal;
	v->cpt_static_prio = tsk->static_prio;
	v->cpt_rt_priority = tsk->rt_priority;
	v->cpt_policy = tsk->policy;
	if (v->cpt_policy != SCHED_NORMAL) {
		eprintk_ctx("scheduler policy is not supported %d/%d(%s)\n", virt_pid(tsk), tsk->pid, tsk->comm);
		cpt_release_buf(ctx);
		return -EINVAL;
	}

	v->cpt_mm = CPT_NULL;
	if (tsk->mm) {
		tobj = lookup_cpt_object(CPT_OBJ_MM, tsk->mm, ctx);
		if (!tobj) BUG();
		v->cpt_mm = tobj->o_pos;
	}
	v->cpt_files = CPT_NULL;
	if (tsk->files) {
		tobj = lookup_cpt_object(CPT_OBJ_FILES, tsk->files, ctx);
		if (!tobj) BUG();
		v->cpt_files = tobj->o_pos;
	}
	v->cpt_fs = CPT_NULL;
	if (tsk->fs) {
		tobj = lookup_cpt_object(CPT_OBJ_FS, tsk->fs, ctx);
		if (!tobj) BUG();
		v->cpt_fs = tobj->o_pos;
	}
	v->cpt_namespace = CPT_NULL;
	if (tsk->namespace) {
		tobj = lookup_cpt_object(CPT_OBJ_NAMESPACE, tsk->namespace, ctx);
		if (!tobj) BUG();
		v->cpt_namespace = tobj->o_pos;

		if (tsk->namespace != current->namespace)
			eprintk_ctx("namespaces are not supported: process %d/%d(%s)\n", virt_pid(tsk), tsk->pid, tsk->comm);
	}
	v->cpt_sysvsem_undo = CPT_NULL;
	if (tsk->sysvsem.undo_list && !tsk->exit_state) {
		tobj = lookup_cpt_object(CPT_OBJ_SYSVSEM_UNDO, tsk->sysvsem.undo_list, ctx);
		if (!tobj) BUG();
		v->cpt_sysvsem_undo = tobj->o_pos;
	}
	v->cpt_sighand = CPT_NULL;
	if (tsk->sighand) {
		tobj = lookup_cpt_object(CPT_OBJ_SIGHAND_STRUCT, tsk->sighand, ctx);
		if (!tobj) BUG();
		v->cpt_sighand = tobj->o_pos;
	}
	v->cpt_sigblocked = cpt_sigset_export(&tsk->blocked);
	v->cpt_sigrblocked = cpt_sigset_export(&tsk->real_blocked);
	v->cpt_sigsuspend_blocked = cpt_sigset_export(&tsk->saved_sigmask);

	v->cpt_pid = virt_pid(tsk);
	v->cpt_tgid = virt_tgid(tsk);
	v->cpt_ppid = 0;
	if (tsk->parent) {
		if (tsk->parent != tsk->real_parent &&
		    !lookup_cpt_object(CPT_OBJ_TASK, tsk->parent, ctx)) {
			eprintk_ctx("task %d/%d(%s) is ptraced from ve0\n", tsk->pid, virt_pid(tsk), tsk->comm);
			cpt_release_buf(ctx);
			return -EBUSY;
		}
		v->cpt_ppid = virt_pid(tsk->parent);
	}
	v->cpt_rppid = tsk->real_parent ? virt_pid(tsk->real_parent) : 0;
	v->cpt_pgrp = virt_pgid(tsk);
	v->cpt_session = virt_sid(tsk);
	v->cpt_old_pgrp = 0;
	if (tsk->signal->tty_old_pgrp)
		v->cpt_old_pgrp = _pid_type_to_vpid(PIDTYPE_PGID, tsk->signal->tty_old_pgrp);
	v->cpt_leader = tsk->group_leader ? virt_pid(tsk->group_leader) : 0;
	v->cpt_set_tid = (unsigned long)tsk->set_child_tid;
	v->cpt_clear_tid = (unsigned long)tsk->clear_child_tid;
	memcpy(v->cpt_comm, tsk->comm, 16);
	v->cpt_user = tsk->user->uid;
	v->cpt_uid = tsk->uid;
	v->cpt_euid = tsk->euid;
	v->cpt_suid = tsk->suid;
	v->cpt_fsuid = tsk->fsuid;
	v->cpt_gid = tsk->gid;
	v->cpt_egid = tsk->egid;
	v->cpt_sgid = tsk->sgid;
	v->cpt_fsgid = tsk->fsgid;
	v->cpt_ngids = 0;
	if (tsk->group_info && tsk->group_info->ngroups != 0) {
		int i = tsk->group_info->ngroups;
		if (i > 32) {
			/* Shame... I did a simplified version and _forgot_
			 * about this. Later, later. */
			eprintk_ctx("too many of groups " CPT_FID "\n", CPT_TID(tsk));
			return -EINVAL;
		}
		v->cpt_ngids = i;
		for (i--; i>=0; i--)
			v->cpt_gids[i] = tsk->group_info->small_block[i];
	}
	memcpy(&v->cpt_ecap, &tsk->cap_effective, 8);
	memcpy(&v->cpt_icap, &tsk->cap_inheritable, 8);
	memcpy(&v->cpt_pcap, &tsk->cap_permitted, 8);
	v->cpt_keepcap = tsk->keep_capabilities;

	v->cpt_did_exec = tsk->did_exec;
	v->cpt_exec_domain = -1;
	v->cpt_thrflags = tsk->thread_info->flags & ~(1<<TIF_FREEZE);
	v->cpt_64bit = 0;
#ifdef CONFIG_X86_64
	/* Clear x86_64 specific flags */
	v->cpt_thrflags &= ~(_TIF_FORK|_TIF_ABI_PENDING|_TIF_IA32);
	if (!(tsk->thread_info->flags & _TIF_IA32)) {
		ctx->tasks64++;
		v->cpt_64bit = 1;
	}
#endif
	v->cpt_thrstatus = tsk->thread_info->status;
	v->cpt_addr_limit = -1;

	v->cpt_personality = tsk->personality;

	for (i=0; i<GDT_ENTRY_TLS_ENTRIES; i++) {
		if (i>=3) {
			eprintk_ctx("too many tls descs\n");
			cpt_release_buf(ctx);
			return -EINVAL;
		}
#ifndef CONFIG_X86_64
		v->cpt_tls[i] = (((u64)tsk->thread.tls_array[i].b)<<32) + tsk->thread.tls_array[i].a;
#else
		v->cpt_tls[i] = tsk->thread.tls_array[i];
#endif
	}

	v->cpt_restart.fn = CPT_RBL_0;
	if (tsk->thread_info->restart_block.fn != current->thread_info->restart_block.fn) {
		if (tsk->thread_info->restart_block.fn != nanosleep_restart
#ifdef CONFIG_X86_64
		    && tsk->thread_info->restart_block.fn != compat_nanosleep_restart
#endif
		    ) {
			eprintk_ctx("unknown restart block %p\n", tsk->thread_info->restart_block.fn);
			cpt_release_buf(ctx);
			return -EINVAL;
		}
		v->cpt_restart.fn = CPT_RBL_NANOSLEEP;
#ifdef CONFIG_X86_64
		if (tsk->thread_info->restart_block.fn == compat_nanosleep_restart)
			v->cpt_restart.fn = CPT_RBL_COMPAT_NANOSLEEP;
#endif
		v->cpt_restart.arg0 = tsk->thread_info->restart_block.arg0;
		v->cpt_restart.arg1 = tsk->thread_info->restart_block.arg1;
		v->cpt_restart.arg2 = tsk->thread_info->restart_block.arg2;
		v->cpt_restart.arg3 = tsk->thread_info->restart_block.arg3;
		if (debug_level > 2) {
			ktime_t e, e1;
			struct timespec now;

			do_posix_clock_monotonic_gettime(&now);
			e = timespec_to_ktime(now);
			e1.tv64 = ((u64)tsk->thread_info->restart_block.arg1 << 32) | (u64) tsk->thread_info->restart_block.arg0;
			e = ktime_sub(e1, e);
			dprintk("cpt " CPT_FID " RBL %ld/%ld %Ld\n", CPT_TID(tsk),
				tsk->thread_info->restart_block.arg1,
				tsk->thread_info->restart_block.arg0, e.tv64);
		}
	}

	v->cpt_it_real_incr = 0;
	v->cpt_it_prof_incr = 0;
	v->cpt_it_virt_incr = 0;
	v->cpt_it_real_value = 0;
	v->cpt_it_prof_value = 0;
	v->cpt_it_virt_value = 0;
	if (thread_group_leader(tsk) && tsk->exit_state == 0) {
		ktime_t rem;

		v->cpt_it_real_incr = ktime_to_ns(tsk->signal->it_real_incr);
		v->cpt_it_prof_incr = tsk->signal->it_prof_incr;
		v->cpt_it_virt_incr = tsk->signal->it_virt_incr;

		rem = hrtimer_get_remaining(&tsk->signal->real_timer);

		if (hrtimer_active(&tsk->signal->real_timer)) {
			if (rem.tv64 <= 0)
				rem.tv64 = NSEC_PER_USEC;
			v->cpt_it_real_value = ktime_to_ns(rem);
			dprintk("cpt itimer " CPT_FID " %Lu\n", CPT_TID(tsk), v->cpt_it_real_value);
		}
		v->cpt_it_prof_value = tsk->signal->it_prof_expires;
		v->cpt_it_virt_value = tsk->signal->it_virt_expires;
	}
	v->cpt_used_math = (tsk_used_math(tsk) != 0);

	if (tsk->notifier) {
		eprintk_ctx("task notifier is in use: process %d/%d(%s)\n", virt_pid(tsk), tsk->pid, tsk->comm);
		cpt_release_buf(ctx);
		return -EINVAL;
	}

	v->cpt_utime = tsk->utime;
	v->cpt_stime = tsk->stime;
	delta = tsk->start_time;
	_set_normalized_timespec(&delta,
			delta.tv_sec - get_exec_env()->init_entry->start_time.tv_sec,
			delta.tv_nsec - get_exec_env()->init_entry->start_time.tv_nsec);
	v->cpt_starttime = cpt_timespec_export(&delta);
	v->cpt_nvcsw = tsk->nvcsw;
	v->cpt_nivcsw = tsk->nivcsw;
	v->cpt_min_flt = tsk->min_flt;
	v->cpt_maj_flt = tsk->maj_flt;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
	v->cpt_cutime = tsk->cutime;
	v->cpt_cstime = tsk->cstime;
	v->cpt_cnvcsw = tsk->cnvcsw;
	v->cpt_cnivcsw = tsk->cnivcsw;
	v->cpt_cmin_flt = tsk->cmin_flt;
	v->cpt_cmaj_flt = tsk->cmaj_flt;

	if (RLIM_NLIMITS > CPT_RLIM_NLIMITS)
		__asm__("undefined\n");

	for (i=0; i<CPT_RLIM_NLIMITS; i++) {
		if (i < RLIM_NLIMITS) {
			v->cpt_rlim_cur[i] = tsk->rlim[i].rlim_cur;
			v->cpt_rlim_max[i] = tsk->rlim[i].rlim_max;
		} else {
			v->cpt_rlim_cur[i] = CPT_NULL;
			v->cpt_rlim_max[i] = CPT_NULL;
		}
	}
#else
	v->cpt_cutime = tsk->signal->cutime;
	v->cpt_cstime = tsk->signal->cstime;
	v->cpt_cnvcsw = tsk->signal->cnvcsw;
	v->cpt_cnivcsw = tsk->signal->cnivcsw;
	v->cpt_cmin_flt = tsk->signal->cmin_flt;
	v->cpt_cmaj_flt = tsk->signal->cmaj_flt;

	if (RLIM_NLIMITS > CPT_RLIM_NLIMITS)
		__asm__("undefined\n");

	for (i=0; i<CPT_RLIM_NLIMITS; i++) {
		if (i < RLIM_NLIMITS) {
			v->cpt_rlim_cur[i] = tsk->signal->rlim[i].rlim_cur;
			v->cpt_rlim_max[i] = tsk->signal->rlim[i].rlim_max;
		} else {
			v->cpt_rlim_cur[i] = CPT_NULL;
			v->cpt_rlim_max[i] = CPT_NULL;
		}
	}
#endif

	if (tsk->mm)
		v->cpt_mm_ub = cpt_lookup_ubc(tsk->mm->mm_ub, ctx);
	else
		v->cpt_mm_ub = CPT_NULL;
	v->cpt_task_ub = cpt_lookup_ubc(tsk->task_bc.task_ub, ctx);
	v->cpt_exec_ub = cpt_lookup_ubc(tsk->task_bc.exec_ub, ctx);
	v->cpt_fork_sub = cpt_lookup_ubc(tsk->task_bc.fork_sub, ctx);

	v->cpt_ptrace_message = tsk->ptrace_message;
	v->cpt_pn_state = tsk->pn_state;
	v->cpt_stopped_state = tsk->stopped_state;
	v->cpt_sigsuspend_state = 0;

#ifndef CONFIG_X86_64
	if (tsk->thread.vm86_info) {
		eprintk_ctx("vm86 task is running\n");
		cpt_release_buf(ctx);
		return -EBUSY;
	}
#endif

	v->cpt_sigpending = cpt_sigset_export(&tsk->pending.signal);

	ctx->write(v, sizeof(*v), ctx);
	cpt_release_buf(ctx);

	cpt_push_object(&saved_obj, ctx);
	dump_kstack(tsk, ctx);
	cpt_pop_object(&saved_obj, ctx);

	cpt_push_object(&saved_obj, ctx);
	err = dump_registers(tsk, ctx);
	cpt_pop_object(&saved_obj, ctx);
	if (err)
		return err;

	if (tsk_used_math(tsk)) {
		cpt_push_object(&saved_obj, ctx);
		dump_fpustate(tsk, ctx);
		cpt_pop_object(&saved_obj, ctx);
	}

	if (tsk->last_siginfo) {
		struct cpt_siginfo_image si;
		cpt_push_object(&saved_obj, ctx);

		si.cpt_next = sizeof(si);
		si.cpt_object = CPT_OBJ_LASTSIGINFO;
		si.cpt_hdrlen = sizeof(si);
		si.cpt_content = CPT_CONTENT_VOID;

		if (encode_siginfo(&si, tsk->last_siginfo))
			return -EINVAL;

		ctx->write(&si, sizeof(si), ctx);
		cpt_pop_object(&saved_obj, ctx);
	}

	if (tsk->sas_ss_size) {
		struct cpt_sigaltstack_image si;
		cpt_push_object(&saved_obj, ctx);

		si.cpt_next = sizeof(si);
		si.cpt_object = CPT_OBJ_SIGALTSTACK;
		si.cpt_hdrlen = sizeof(si);
		si.cpt_content = CPT_CONTENT_VOID;

		si.cpt_stack = tsk->sas_ss_sp;
		si.cpt_stacksize = tsk->sas_ss_size;

		ctx->write(&si, sizeof(si), ctx);
		cpt_pop_object(&saved_obj, ctx);
	}

	dump_sigqueue(&tsk->pending, ctx);

	last_thread = 1;
	read_lock(&tasklist_lock);
	do {
		task_t * next = next_thread(tsk);
		if (next != tsk && !thread_group_leader(next))
			last_thread = 0;
	} while (0);
	read_unlock(&tasklist_lock);

	if (last_thread) {
		task_t *prev_tsk;
		int err;
		loff_t pos = ctx->file->f_pos;

		cpt_push_object(&saved_obj, ctx);
		err = dump_one_signal_struct(tg_obj, ctx);
		cpt_pop_object(&saved_obj, ctx);
		if (err)
			return err;

		prev_tsk = tsk;
		for (;;) {
			if (prev_tsk->tgid == tsk->tgid) {
				loff_t tg_pos;

				tg_pos = obj->o_pos + offsetof(struct cpt_task_image, cpt_signal);
				ctx->pwrite(&pos, sizeof(pos), ctx, tg_pos);
				if (thread_group_leader(prev_tsk))
					break;
			}

			if (obj->o_list.prev == &ctx->object_array[CPT_OBJ_TASK]) {
				eprintk_ctx("bug: thread group leader is lost\n");
				return -EINVAL;
			}

			obj = list_entry(obj->o_list.prev, cpt_object_t, o_list);
			prev_tsk = obj->o_obj;
		}
	}

	cpt_close_object(ctx);
	return 0;
}

int cpt_dump_tasks(struct cpt_context *ctx)
{
	cpt_object_t *obj;

	cpt_open_section(ctx, CPT_SECT_TASKS);

	for_each_object(obj, CPT_OBJ_TASK) {
		int err;

		if ((err = dump_one_process(obj, ctx)) != 0)
			return err;
	}

	cpt_close_section(ctx);
	return 0;
}

int cpt_collect_signals(cpt_context_t *ctx)
{
	cpt_object_t *obj;

	/* Collect process fd sets */
	for_each_object(obj, CPT_OBJ_TASK) {
		task_t *tsk = obj->o_obj;
		if (tsk->signal && !list_empty(&tsk->signal->posix_timers)) {
			eprintk_ctx("task %d/%d(%s) uses posix timers\n", tsk->pid, virt_pid(tsk), tsk->comm);
			return -EBUSY;
		}
		if (tsk->signal && cpt_object_add(CPT_OBJ_SIGNAL_STRUCT, tsk->signal, ctx) == NULL)
			return -ENOMEM;
		if (tsk->sighand && cpt_object_add(CPT_OBJ_SIGHAND_STRUCT, tsk->sighand, ctx) == NULL)
			return -ENOMEM;
	}
	return 0;
}


static int dump_one_sighand_struct(cpt_object_t *obj, struct cpt_context *ctx)
{
	struct sighand_struct *sig = obj->o_obj;
	struct cpt_sighand_image *v = cpt_get_buf(ctx);
	int i;

	cpt_open_object(obj, ctx);

	v->cpt_next = CPT_NULL;
	v->cpt_object = CPT_OBJ_SIGHAND_STRUCT;
	v->cpt_hdrlen = sizeof(*v);
	v->cpt_content = CPT_CONTENT_ARRAY;

	ctx->write(v, sizeof(*v), ctx);
	cpt_release_buf(ctx);

	for (i=0; i< _NSIG; i++) {
		if (sig->action[i].sa.sa_handler != SIG_DFL) {
			loff_t saved_obj;
			struct cpt_sighandler_image *o = cpt_get_buf(ctx);

			cpt_push_object(&saved_obj, ctx);
			cpt_open_object(NULL, ctx);

			o->cpt_next = CPT_NULL;
			o->cpt_object = CPT_OBJ_SIGHANDLER;
			o->cpt_hdrlen = sizeof(*o);
			o->cpt_content = CPT_CONTENT_VOID;

			o->cpt_signo = i;
			o->cpt_handler = (unsigned long)sig->action[i].sa.sa_handler;
			o->cpt_restorer = (unsigned long)sig->action[i].sa.sa_restorer;
			o->cpt_flags = sig->action[i].sa.sa_flags;
			memcpy(&o->cpt_mask, &sig->action[i].sa.sa_mask, 8);
			ctx->write(o, sizeof(*o), ctx);
			cpt_release_buf(ctx);
			cpt_close_object(ctx);
			cpt_pop_object(&saved_obj, ctx);
		}
	}

	cpt_close_object(ctx);
	return 0;
}

int cpt_dump_sighand(struct cpt_context *ctx)
{
	cpt_object_t *obj;

	cpt_open_section(ctx, CPT_SECT_SIGHAND_STRUCT);

	for_each_object(obj, CPT_OBJ_SIGHAND_STRUCT) {
		int err;

		if ((err = dump_one_sighand_struct(obj, ctx)) != 0)
			return err;
	}

	cpt_close_section(ctx);
	return 0;
}
