/*
 *
 *  kernel/cpt/rst_net.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/socket.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <linux/rtnetlink.h>
#include <linux/ve.h>
#include <linux/ve_proto.h>
#include <net/route.h>
#include <net/ip_fib.h>
#include <net/addrconf.h>

#include "cpt_obj.h"
#include "cpt_context.h"
#include "cpt_kernel.h"
#include "cpt_net.h"

#include "cpt_syscalls.h"

extern struct in_ifaddr *inet_alloc_ifa(void);
extern int inet_insert_ifa(struct in_ifaddr *ifa);

int rst_restore_ifaddr(struct cpt_context *ctx)
{
	int err;
	loff_t sec = ctx->sections[CPT_SECT_NET_IFADDR];
	loff_t endsec;
	struct cpt_section_hdr h;
	struct cpt_ifaddr_image di;
	struct net_device *dev;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_NET_IFADDR || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	endsec = sec + h.cpt_next;
	sec += h.cpt_hdrlen;
	while (sec < endsec) {
		int cindex = -1;
		int err;
		err = rst_get_object(CPT_OBJ_NET_IFADDR, sec, &di, ctx);
		if (err)
			return err;
		if (di.cpt_index == ctx->lo_index_old)
			cindex = ctx->lo_index;
		else if (di.cpt_index == ctx->venet_index_old)
			cindex = ctx->venet_index;
		if (cindex <= 0)
			eprintk_ctx("unknown ifaddr for %d\n", di.cpt_index);
		rtnl_lock();
		dev = __dev_get_by_index(cindex);
		if (dev && di.cpt_family == AF_INET) {
			struct in_device *in_dev;
			struct in_ifaddr *ifa;
			if ((in_dev = __in_dev_get_rtnl(dev)) == NULL)
				in_dev = inetdev_init(dev);
			ifa = inet_alloc_ifa();
			if (ifa) {
				ifa->ifa_local = di.cpt_address[0];
				ifa->ifa_address = di.cpt_peer[0];
				ifa->ifa_broadcast = di.cpt_broadcast[0];
				ifa->ifa_prefixlen = di.cpt_masklen;
				ifa->ifa_mask = inet_make_mask(ifa->ifa_prefixlen);
				ifa->ifa_flags = di.cpt_flags;
				ifa->ifa_scope = di.cpt_scope;
				memcpy(ifa->ifa_label, di.cpt_label, IFNAMSIZ);
				in_dev_hold(in_dev);
				ifa->ifa_dev   = in_dev;
				err = inet_insert_ifa(ifa);
				if (err && err != -EEXIST) {
					rtnl_unlock();
					eprintk_ctx("add ifaddr err %d for %d %s\n", err, di.cpt_index, di.cpt_label);
					return err;
				}
			}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
		} else if (dev && di.cpt_family == AF_INET6) {
			err = inet6_addr_add(dev->ifindex,
					     (struct in6_addr *)di.cpt_address,
					     di.cpt_masklen);
			if (err && err != -EEXIST) {
				rtnl_unlock();
				eprintk_ctx("add ifaddr err %d for %d %s\n", err, di.cpt_index, di.cpt_label);
				return err;
			}
#endif
		} else {
			rtnl_unlock();
			eprintk_ctx("unknown ifaddr 2 for %d\n", di.cpt_index);
			return -EINVAL;
		}
		rtnl_unlock();
		sec += di.cpt_next;
	}
	return 0;
}

static int rewrite_rtmsg(struct nlmsghdr *nlh, struct cpt_context *ctx)
{
	int min_len = NLMSG_LENGTH(sizeof(struct rtmsg));
	struct rtmsg *rtm = NLMSG_DATA(nlh);
	int idx = -1;
	__u32 prefix0 = 0;

	if (nlh->nlmsg_len > min_len) {
		int attrlen = nlh->nlmsg_len - NLMSG_ALIGN(min_len);
		struct rtattr *rta = (void*)nlh + NLMSG_ALIGN(min_len);

		while (RTA_OK(rta, attrlen)) {
			if (rta->rta_type == RTA_OIF) {
				idx = *(int*)RTA_DATA(rta);
				if (idx == ctx->lo_index_old)
					idx = ctx->lo_index;
				else if (idx == ctx->venet_index_old)
					idx = ctx->venet_index;
				else {
					eprintk_ctx("unknown iface %d\n", idx);
					return -ENODEV;
				}
				*(int*)RTA_DATA(rta) = idx;
			} else if (rta->rta_type == RTA_DST) {
				prefix0 = *(__u32*)RTA_DATA(rta);
			}
			rta = RTA_NEXT(rta, attrlen);
		}
	}
#if defined(CONFIG_IPV6) || defined (CONFIG_IPV6_MODULE)
	if (rtm->rtm_family == AF_INET6) {
		if (rtm->rtm_type == RTN_LOCAL)
			return 2;
		if (rtm->rtm_flags & RTM_F_CLONED)
			return 2;
		if (rtm->rtm_protocol == RTPROT_UNSPEC ||
		    rtm->rtm_protocol == RTPROT_RA ||
		    rtm->rtm_protocol == RTPROT_REDIRECT ||
		    rtm->rtm_protocol == RTPROT_KERNEL)
			return 2;
		if (rtm->rtm_protocol == RTPROT_BOOT &&
		    ((rtm->rtm_dst_len == 8 && prefix0 == htonl(0xFF000000)) ||
		     (rtm->rtm_dst_len == 64 && prefix0 == htonl(0xFE800000))))
			return 2;
	}
#endif
	return rtm->rtm_protocol == RTPROT_KERNEL;
}

int rst_restore_route(struct cpt_context *ctx)
{
	int err;
	struct socket *sock;
	struct msghdr msg;
	struct iovec iov;
	struct sockaddr_nl nladdr;
	mm_segment_t oldfs;
	loff_t sec = ctx->sections[CPT_SECT_NET_ROUTE];
	loff_t endsec;
	struct cpt_section_hdr h;
	struct cpt_object_hdr v;
	char *pg;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_NET_ROUTE || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	if (h.cpt_hdrlen >= h.cpt_next)
		return 0;

	sec += h.cpt_hdrlen;
	err = rst_get_object(CPT_OBJ_NET_ROUTE, sec, &v, ctx);
	if (err < 0)
		return err;

	err = sock_create_kern(AF_NETLINK, SOCK_DGRAM, NETLINK_ROUTE, &sock);
	if (err)
		return err;

	pg = (char*)__get_free_page(GFP_KERNEL);
	if (pg == NULL) {
		err = -ENOMEM;
		goto out_sock;
	}

	memset(&nladdr, 0, sizeof(nladdr));
	nladdr.nl_family = AF_NETLINK;

	endsec = sec + v.cpt_next;
	sec += v.cpt_hdrlen;

	while (sec < endsec) {
		struct nlmsghdr *n;
		struct nlmsghdr nh;
		int kernel_flag;

		err = ctx->pread(&nh, sizeof(nh), ctx, sec);
		if (err)
			goto out_sock_pg;
		if (nh.nlmsg_len > PAGE_SIZE) {
			err = -EINVAL;
			goto out_sock_pg;
		}
		err = ctx->pread(pg, nh.nlmsg_len, ctx, sec);
		if (err)
			goto out_sock_pg;

		n = (struct nlmsghdr*)pg;
		n->nlmsg_flags = NLM_F_REQUEST|NLM_F_APPEND|NLM_F_CREATE;

		err = rewrite_rtmsg(n, ctx);
		if (err < 0)
			goto out_sock_pg;
		kernel_flag = err;

		if (kernel_flag == 2)
			goto do_next;

		iov.iov_base=n;
		iov.iov_len=nh.nlmsg_len;
		msg.msg_name=&nladdr;
		msg.msg_namelen=sizeof(nladdr);
		msg.msg_iov=&iov;
		msg.msg_iovlen=1;
		msg.msg_control=NULL;
		msg.msg_controllen=0;
		msg.msg_flags=MSG_DONTWAIT;

		oldfs = get_fs(); set_fs(KERNEL_DS);
		err = sock_sendmsg(sock, &msg, nh.nlmsg_len);
		set_fs(oldfs);

		if (err < 0)
			goto out_sock_pg;
		err = 0;

		iov.iov_base=pg;
		iov.iov_len=PAGE_SIZE;

		oldfs = get_fs(); set_fs(KERNEL_DS);
		err = sock_recvmsg(sock, &msg, PAGE_SIZE, MSG_DONTWAIT);
		set_fs(oldfs);
		if (err != -EAGAIN) {
			if (err == NLMSG_LENGTH(sizeof(struct nlmsgerr)) &&
			    n->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr *e = NLMSG_DATA(n);
				if (e->error != -EEXIST || !kernel_flag)
					eprintk_ctx("NLMERR: %d\n", e->error);
			} else {
				eprintk_ctx("Res: %d %d\n", err, n->nlmsg_type);
			}
		}
do_next:
		err = 0;
		sec += NLMSG_ALIGN(nh.nlmsg_len);
	}

out_sock_pg:
	free_page((unsigned long)pg);
out_sock:
	sock_release(sock);
	return err;
}

int rst_resume_network(struct cpt_context *ctx)
{
	struct ve_struct *env;

	env = get_ve_by_id(ctx->ve_id);
	if (!env)
		return -ESRCH;
	env->disable_net = 0;
	put_ve(env);
	return 0;
}

int rst_restore_netdev(struct cpt_context *ctx)
{
	int err;
	loff_t sec = ctx->sections[CPT_SECT_NET_DEVICE];
	loff_t endsec;
	struct cpt_section_hdr h;
	struct cpt_netdev_image di;
	struct net_device *dev;

	get_exec_env()->disable_net = 1;

	dev = __dev_get_by_name("lo");
	if (!dev) {
		eprintk_ctx("cannot find loopback netdevice\n");
		return -EINVAL;
	}
	ctx->lo_index = dev->ifindex;
	ctx->lo_index_old = -1;
	dev = __dev_get_by_name("venet0");
	if (!dev) {
		eprintk_ctx("cannot find venet0 netdevice\n");
		return -EINVAL;
	}
	ctx->venet_index = dev->ifindex;
	ctx->venet_index_old = -1;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_NET_DEVICE || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	endsec = sec + h.cpt_next;
	sec += h.cpt_hdrlen;
	while (sec < endsec) {
		int err;
		err = rst_get_object(CPT_OBJ_NET_DEVICE, sec, &di, ctx);
		if (err)
			return err;
		if (strcmp(di.cpt_name, "lo") == 0) {
			ctx->lo_index_old = di.cpt_index;
		} else if (strcmp(di.cpt_name, "venet0") == 0) {
			ctx->venet_index_old = di.cpt_index;
		} else {
			eprintk_ctx("unknown interface %s\n", di.cpt_name);
		}
		dev = __dev_get_by_name(di.cpt_name);
		if (dev) {
			if (di.cpt_flags^dev->flags) {
				rtnl_lock();
				err = dev_change_flags(dev, di.cpt_flags);
				rtnl_unlock();
				if (err)
					eprintk_ctx("dev_change_flags err: %d\n", err);
			}
		} else {
			eprintk_ctx("unknown interface 2 %s\n", di.cpt_name);
		}
		sec += di.cpt_next;
	}
	return 0;
}

static int dumpfn(void *arg)
{
	int i;
	int *pfd = arg;
	char *argv[] = { "iptables-restore", "-c", NULL };

	if (pfd[0] != 0)
		sc_dup2(pfd[0], 0);

	for (i=1; i<current->files->fdt->max_fds; i++)
		sc_close(i);

	module_put(THIS_MODULE);

	set_fs(KERNEL_DS);
	i = sc_execve("/sbin/iptables-restore", argv, NULL);
	eprintk("failed to exec /sbin/iptables-restore: %d\n", i);
	return -1;
}

static int rst_restore_iptables(struct cpt_context * ctx)
{
	int err;
	int pfd[2];
	struct file *f;
	struct cpt_object_hdr v;
	int n;
	struct cpt_section_hdr h;
	loff_t sec = ctx->sections[CPT_SECT_NET_IPTABLES];
	loff_t end;
	int pid;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_NET_IPTABLES || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	if (h.cpt_hdrlen == h.cpt_next)
		return 0;
	if (h.cpt_hdrlen > h.cpt_next)
		return -EINVAL;
	sec += h.cpt_hdrlen;
	err = rst_get_object(CPT_OBJ_NAME, sec, &v, ctx);
	if (err < 0)
		return err;

	err = sc_pipe(pfd);
	if (err < 0)
		return err;
	pid = err = local_kernel_thread(dumpfn, (void*)pfd, SIGCHLD, 0);
	if (err < 0)
		goto out;
	f = fget(pfd[1]);
	sc_close(pfd[1]);
	sc_close(pfd[0]);

	ctx->file->f_pos = sec + v.cpt_hdrlen;
	end = sec + v.cpt_next;
	do {
		char *p;
		char buf[16];
		mm_segment_t oldfs;

		n = end - ctx->file->f_pos;
		if (n > sizeof(buf))
			n = sizeof(buf);

		if (ctx->read(buf, n, ctx))
			break;
		if ((p = memchr(buf, 0, n)) != NULL)
			n = p - buf;
		oldfs = get_fs(); set_fs(KERNEL_DS);
		f->f_op->write(f, buf, n, &f->f_pos);
		set_fs(oldfs);
	} while (ctx->file->f_pos < end);

	fput(f);

	clear_tsk_thread_flag(current,TIF_SIGPENDING);

	if ((err = sc_waitx(pid, 0)) < 0)
		eprintk_ctx("wait4: %d\n", err);

	return 0;

out:
	if (pfd[1] >= 0)
		sc_close(pfd[1]);
	if (pfd[0] >= 0)
		sc_close(pfd[0]);
	return err;
}

int rst_restore_net(struct cpt_context *ctx)
{
	int err;

	err = rst_restore_netdev(ctx);
	if (!err)
		err = rst_restore_ifaddr(ctx);
	if (!err)
		err = rst_restore_route(ctx);
	if (!err)
		err = rst_restore_iptables(ctx);
	if (!err)
		err = rst_restore_ip_conntrack(ctx);
	return err;
}
