/*
 *
 *  kernel/cpt/rst_process.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/virtinfo.h>
#include <linux/kmem_cache.h>
#include <linux/errno.h>
#include <linux/pagemap.h>
#include <linux/ptrace.h>
#include <linux/tty.h>
#include <asm/desc.h>
#include <asm/unistd.h>

#include <ub/beancounter.h>
#include <ub/ub_misc.h>

#include "cpt_obj.h"
#include "cpt_context.h"
#include "cpt_files.h"
#include "cpt_mm.h"
#include "cpt_ubc.h"
#include "cpt_process.h"
#include "cpt_kernel.h"

#ifdef CONFIG_X86_64

#define _TIF_RESUME (1<<22)

#define SYSCALL_NR(regs) ((regs)->orig_rax)
#define SYSCALL_RETVAL(regs) ((regs)->rax)
#define SYSCALL_PC(regs) ((regs)->rip)

#define ESP(tsk) (tsk)->thread.rsp

#define __NR32_restart_syscall	0
#define __NR32_rt_sigtimedwait	177
#define __NR32_pause		29
#define __NR32_futex		240

#define syscall_is(tsk,regs,name) ((!((tsk)->thread_info->flags&_TIF_IA32) && \
				    SYSCALL_NR(regs) == __NR_##name) || \
				   (((tsk)->thread_info->flags&_TIF_IA32) && \
				    SYSCALL_NR(regs) == __NR32_##name))
#else

#define SYSCALL_NR(regs) ((regs)->orig_eax)
#define SYSCALL_RETVAL(regs) ((regs)->eax)
#define SYSCALL_PC(regs) ((regs)->eip)

#define ESP(tsk) (tsk)->thread.esp

#define syscall_is(tsk,regs,name) (SYSCALL_NR(regs) == __NR_##name)

#undef task_pt_regs
#define task_pt_regs(t) ((struct pt_regs *)((t)->thread.esp0) - 1)

#endif

static void decode_siginfo(siginfo_t *info, struct cpt_siginfo_image *si)
{
	memset(info, 0, sizeof(*info));
	switch(si->cpt_code & __SI_MASK) {
	case __SI_TIMER:
		info->si_tid = si->cpt_pid;
		info->si_overrun = si->cpt_uid;
		info->_sifields._timer._sigval.sival_ptr = cpt_ptr_import(si->cpt_sigval);
		info->si_sys_private = si->cpt_utime;
		break;
	case __SI_POLL:
		info->si_band = si->cpt_pid;
		info->si_fd = si->cpt_uid;
		break;
	case __SI_FAULT:
		info->si_addr = cpt_ptr_import(si->cpt_sigval);
#ifdef __ARCH_SI_TRAPNO
		info->si_trapno = si->cpt_pid;
#endif
		break;
	case __SI_CHLD:
		info->si_pid = si->cpt_pid;
		info->si_uid = si->cpt_uid;
		info->si_status = si->cpt_sigval;
		info->si_stime = si->cpt_stime;
		info->si_utime = si->cpt_utime;
		break;
	case __SI_KILL:
	case __SI_RT:
	case __SI_MESGQ:
	default:
		info->si_pid = si->cpt_pid;
		info->si_uid = si->cpt_uid;
		info->si_ptr = cpt_ptr_import(si->cpt_sigval);
		break;
	}
	info->si_signo = si->cpt_signo;
	info->si_errno = si->cpt_errno;
	info->si_code = si->cpt_code;
}

static int restore_sigqueue(task_t *tsk,
			    struct sigpending *queue, unsigned long start,
			    unsigned long end)
{
	while (start < end) {
		struct cpt_siginfo_image *si = (struct cpt_siginfo_image *)start;
		if (si->cpt_object == CPT_OBJ_SIGINFO) {
			struct sigqueue *q = NULL;
			struct user_struct *up;
			up = alloc_uid(si->cpt_user);
			if (!up)
				return -ENOMEM;
			q = kmem_cache_alloc(sigqueue_cachep, GFP_ATOMIC);
			if (!q) {
				free_uid(up);
				return -ENOMEM;
			}
			if (ub_siginfo_charge(q, get_exec_ub())) {
				kmem_cache_free(sigqueue_cachep, q);
				free_uid(up);
				return -ENOMEM;
			}

			INIT_LIST_HEAD(&q->list);
			/* Preallocated elements (posix timers) are not
			 * supported yet. It is safe to replace them with
			 * a private one. */
			q->flags = 0;
			q->user = up;
			atomic_inc(&q->user->sigpending);

			decode_siginfo(&q->info, si);
			list_add_tail(&q->list, &queue->list);
		}
		start += si->cpt_next;
	}
	return 0;
}

int rst_process_linkage(cpt_context_t *ctx)
{
	cpt_object_t *obj;

	for_each_object(obj, CPT_OBJ_TASK) {
		task_t *tsk = obj->o_obj;
		struct cpt_task_image *ti = obj->o_image;

		if (tsk == NULL) {
			eprintk_ctx("task %u(%s) is missing\n", ti->cpt_pid, ti->cpt_comm);
			return -EINVAL;
		}

		if (virt_pgid(tsk) != ti->cpt_pgrp) {
			int pid;

			if ((pid = vpid_to_pid(ti->cpt_pgrp)) < 0) {
				eprintk_ctx("illegal PGRP " CPT_FID "\n", CPT_TID(tsk));
				return -EINVAL;
			}

			write_lock_irq(&tasklist_lock);
			detach_pid(tsk, PIDTYPE_PGID);
			tsk->signal->pgrp = pid;
			set_virt_pgid(tsk, ti->cpt_pgrp);
			if (thread_group_leader(tsk))
				attach_pid(tsk, PIDTYPE_PGID, pid);
			write_unlock_irq(&tasklist_lock);
		}
		if (virt_sid(tsk) != ti->cpt_session) {
			int pid;

			if ((pid = vpid_to_pid(ti->cpt_session)) < 0) {
				eprintk_ctx("illegal SID " CPT_FID "\n", CPT_TID(tsk));
				return -EINVAL;
			}

			write_lock_irq(&tasklist_lock);
			detach_pid(tsk, PIDTYPE_SID);
			tsk->signal->session = pid;
			set_virt_sid(tsk, ti->cpt_session);
			if (thread_group_leader(tsk))
				attach_pid(tsk, PIDTYPE_SID, pid);
			write_unlock_irq(&tasklist_lock);
		}
		if (ti->cpt_old_pgrp > 0 && tsk->signal->tty_old_pgrp == 0) {
			int pid;

			if ((pid = vpid_to_pid(ti->cpt_old_pgrp)) < 0) {
				eprintk_ctx("illegal OLD_PGRP " CPT_FID "\n", CPT_TID(tsk));
				return -EINVAL;
			}

			tsk->signal->tty_old_pgrp = pid;
		}
	}

	return 0;
}

static int restore_one_signal_struct(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	int err;
	struct cpt_signal_image *si = cpt_get_buf(ctx);

	current->signal->tty = NULL;

	err = rst_get_object(CPT_OBJ_SIGNAL_STRUCT, ti->cpt_signal, si, ctx);
	if (err) {
		cpt_release_buf(ctx);
		return err;
	}

	if (virt_pgid(current) != si->cpt_pgrp) {
		int err;
		int pid = 0;

		if (si->cpt_pgrp_type == CPT_PGRP_ORPHAN) {
			pid = alloc_pidmap();
			if (pid < 0) {
				cpt_release_buf(ctx);
				return -EINVAL;
			}
			if ((err = alloc_vpid(pid, si->cpt_pgrp)) < 0) {
				free_pidmap(pid);
				pid = 0;
				if (err != -EEXIST) {
					cpt_release_buf(ctx);
					return err;
				}
			}
		}
		if (pid ||
		    (pid = vpid_to_pid(si->cpt_pgrp)) > 0) {
			write_lock_irq(&tasklist_lock);
			detach_pid(current, PIDTYPE_PGID);
			current->signal->pgrp = pid;
			set_virt_pgid(current, si->cpt_pgrp);
			if (thread_group_leader(current))
				attach_pid(current, PIDTYPE_PGID, pid);
			write_unlock_irq(&tasklist_lock);
		}
	}

	current->signal->tty_old_pgrp = 0;
	if ((int)si->cpt_old_pgrp > 0) {
		if (si->cpt_old_pgrp_type == CPT_PGRP_STRAY) {
			current->signal->tty_old_pgrp = alloc_pidmap();
			if (current->signal->tty_old_pgrp < 0) {
				eprintk_ctx("failed to allocate stray tty_old_pgrp\n");
				cpt_release_buf(ctx);
				return -EINVAL;
			}
			free_pidmap(current->signal->tty_old_pgrp);
		} else {
			current->signal->tty_old_pgrp = vpid_to_pid(si->cpt_old_pgrp);
			if (current->signal->tty_old_pgrp < 0) {
				dprintk_ctx("forward old tty PGID\n");
				current->signal->tty_old_pgrp = 0;
			}
		}
	}

	if (virt_sid(current) != si->cpt_session) {
		int err;
		int pid = 0;

		if (si->cpt_session_type == CPT_PGRP_ORPHAN) {
			pid = alloc_pidmap();
			if (pid < 0) {
				cpt_release_buf(ctx);
				return -EINVAL;
			}
			if ((err = alloc_vpid(pid, si->cpt_session)) < 0) {
				free_pidmap(pid);
				pid = 0;
				if (err != -EEXIST) {
					cpt_release_buf(ctx);
					return err;
				}
			}
		}
		if (pid ||
		    (pid = vpid_to_pid(si->cpt_session)) > 0) {
			write_lock_irq(&tasklist_lock);
			detach_pid(current, PIDTYPE_SID);
			set_virt_sid(current, si->cpt_session);
			current->signal->session = pid;
			if (thread_group_leader(current))
				attach_pid(current, PIDTYPE_SID, pid);
			write_unlock_irq(&tasklist_lock);
		}
	}

	cpt_sigset_import(&current->signal->shared_pending.signal, si->cpt_sigpending);
	current->signal->leader = si->cpt_leader;
	if (si->cpt_ctty != CPT_NULL) {
		cpt_object_t *obj = lookup_cpt_obj_bypos(CPT_OBJ_TTY, si->cpt_ctty, ctx);
		if (obj) {
			struct tty_struct *tty = obj->o_obj;
			if (tty->session == 0 || tty->session == current->signal->session) {
				tty->session = current->signal->session;
				current->signal->tty = tty;
			} else {
				wprintk_ctx("tty session mismatch\n");
			}
		}
	}

	if (si->cpt_curr_target)
		current->signal->curr_target = find_task_by_pid_ve(si->cpt_curr_target);
	current->signal->flags = 0;
	if (si->cpt_group_exit)
		current->signal->flags |= SIGNAL_GROUP_EXIT;
	current->signal->group_exit_code = si->cpt_group_exit_code;
	if (si->cpt_group_exit_task) {
		current->signal->group_exit_task = find_task_by_pid_ve(si->cpt_group_exit_task);
		if (current->signal->group_exit_task == NULL) {
			eprintk_ctx("oops, group_exit_task=NULL, pid=%u\n", si->cpt_group_exit_task);
			cpt_release_buf(ctx);
			return -EINVAL;
		}
	}
	current->signal->notify_count = si->cpt_notify_count;
	current->signal->group_stop_count = si->cpt_group_stop_count;

	if (si->cpt_next > si->cpt_hdrlen) {
		char *buf = kmalloc(si->cpt_next - si->cpt_hdrlen, GFP_KERNEL);
		if (buf == NULL) {
			cpt_release_buf(ctx);
			return -ENOMEM;
		}
		err = ctx->pread(buf, si->cpt_next - si->cpt_hdrlen, ctx,
				 ti->cpt_signal + si->cpt_hdrlen);
		if (err) {
			kfree(buf);
			cpt_release_buf(ctx);
			return err;
		}
		restore_sigqueue(current,
				 &current->signal->shared_pending, (unsigned long)buf,
				 (unsigned long)buf + si->cpt_next - si->cpt_hdrlen);
		kfree(buf);
	}
	cpt_release_buf(ctx);
	return 0;
}

int restore_one_sighand_struct(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	int err;
	struct cpt_sighand_image si;
	int i;
	loff_t pos, endpos;

	err = rst_get_object(CPT_OBJ_SIGHAND_STRUCT, ti->cpt_sighand, &si, ctx);
	if (err)
		return err;

	for (i=0; i<_NSIG; i++) {
		current->sighand->action[i].sa.sa_handler = SIG_DFL;
		current->sighand->action[i].sa.sa_restorer = 0;
		current->sighand->action[i].sa.sa_flags = SA_ONESHOT | SA_NOMASK;
		memset(&current->sighand->action[i].sa.sa_mask, 0, sizeof(sigset_t));
	}

	pos = ti->cpt_sighand + si.cpt_hdrlen;
	endpos = ti->cpt_sighand + si.cpt_next;
	while (pos < endpos) {
		struct cpt_sighandler_image shi;

		err = rst_get_object(CPT_OBJ_SIGHANDLER, pos, &shi, ctx);
		if (err)
			return err;
		current->sighand->action[shi.cpt_signo].sa.sa_handler = (void*)(unsigned long)shi.cpt_handler;
		current->sighand->action[shi.cpt_signo].sa.sa_restorer = (void*)(unsigned long)shi.cpt_restorer;
		current->sighand->action[shi.cpt_signo].sa.sa_flags = shi.cpt_flags;
		cpt_sigset_import(&current->sighand->action[shi.cpt_signo].sa.sa_mask, shi.cpt_mask);
		pos += shi.cpt_next;
	}

	return 0;
}


__u32 rst_signal_flag(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	__u32 flag = 0;

	if (lookup_cpt_obj_bypos(CPT_OBJ_SIGNAL_STRUCT, ti->cpt_signal, ctx))
		flag |= CLONE_THREAD;
	if (ti->cpt_sighand == CPT_NULL ||
	    lookup_cpt_obj_bypos(CPT_OBJ_SIGHAND_STRUCT, ti->cpt_sighand, ctx))
		flag |= CLONE_SIGHAND;
	return flag;
}

int rst_signal_complete(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	int err;
	cpt_object_t *obj;

	if (ti->cpt_signal == CPT_NULL || ti->cpt_sighand == CPT_NULL) {
		return -EINVAL;
	}

	obj = lookup_cpt_obj_bypos(CPT_OBJ_SIGHAND_STRUCT, ti->cpt_sighand, ctx);
	if (obj) {
		struct sighand_struct *sig = current->sighand;
		if (obj->o_obj != sig) {
			return -EINVAL;
		}
	} else {
		obj = cpt_object_add(CPT_OBJ_SIGHAND_STRUCT, current->sighand, ctx);
		if (obj == NULL)
			return -ENOMEM;
		cpt_obj_setpos(obj, ti->cpt_sighand, ctx);
		err = restore_one_sighand_struct(ti, ctx);
		if (err)
			return err;
	}


	obj = lookup_cpt_obj_bypos(CPT_OBJ_SIGNAL_STRUCT, ti->cpt_signal, ctx);
	if (obj) {
		struct signal_struct *sig = current->signal;
		if (obj->o_obj != sig) {
			return -EINVAL;
		}
		if (current->signal) {
			set_virt_pgid(current, pid_type_to_vpid(PIDTYPE_PGID, current->signal->pgrp));
			set_virt_sid(current, pid_type_to_vpid(PIDTYPE_SID, current->signal->session));
		}
	} else {
		obj = cpt_object_add(CPT_OBJ_SIGNAL_STRUCT, current->signal, ctx);
		if (obj == NULL)
			return -ENOMEM;
		cpt_obj_setpos(obj, ti->cpt_signal, ctx);
		err = restore_one_signal_struct(ti, ctx);
		if (err)
			return err;
	}

	return 0;
}

static u32 decode_segment(u32 segid)
{
	if (segid == CPT_SEG_ZERO)
		return 0;

	/* TLS descriptors */
	if (segid <= CPT_SEG_TLS3)
		return ((GDT_ENTRY_TLS_MIN + segid-CPT_SEG_TLS1)<<3) + 3;

	/* LDT descriptor, it is just an index to LDT array */
	if (segid >= CPT_SEG_LDT)
		return ((segid - CPT_SEG_LDT) << 3) | 7;

	/* Check for one of standard descriptors */
#ifdef CONFIG_X86_64
	if (segid == CPT_SEG_USER32_DS)
		return __USER32_DS;
	if (segid == CPT_SEG_USER32_CS)
		return __USER32_CS;
	if (segid == CPT_SEG_USER64_DS)
		return __USER_DS;
	if (segid == CPT_SEG_USER64_CS)
		return __USER_CS;
#else
	if (segid == CPT_SEG_USER32_DS)
		return __USER_DS;
	if (segid == CPT_SEG_USER32_CS)
		return __USER_CS;
#endif
	wprintk("Invalid segment reg %d\n", segid);
	return 0;
}

unsigned long rct(unsigned long *child_tids)
{
	dprintk("rct: " CPT_FID "\n", CPT_TID(current));
	current->clear_child_tid = (void*)child_tids[0];
	current->set_child_tid = (void*)child_tids[1];
	module_put(THIS_MODULE);
	return (unsigned long)(child_tids+2);
}

unsigned long rlsi(void)
{
	int signr;
	siginfo_t *info = current->last_siginfo;
	struct pt_regs *regs = task_pt_regs(current);
	struct k_sigaction *ka;
	int ptrace_id;

	dprintk("rlsi: " CPT_FID "\n", CPT_TID(current));

	spin_lock_irq(&current->sighand->siglock);
	current->last_siginfo = NULL;
	recalc_sigpending();

	ptrace_id = current->pn_state;
	clear_pn_state(current);

	switch (ptrace_id) {
	case PN_STOP_TF:
	case PN_STOP_TF_RT:
		/* frame_*signal */
		dprintk("SIGTRAP %u/%u(%s) %u/%u %u %ld %lu %lu\n",
		       virt_pid(current), current->pid, current->comm,
		       info->si_signo, info->si_code,
		       current->exit_code, SYSCALL_NR(regs),
		       current->ptrace, current->ptrace_message);
		goto out;
	case PN_STOP_ENTRY:
	case PN_STOP_LEAVE:
		/* do_syscall_trace */
		spin_unlock_irq(&current->sighand->siglock);
		dprintk("ptrace do_syscall_trace: %d %d\n", ptrace_id, current->exit_code);
		if (current->exit_code) {
			send_sig(current->exit_code, current, 1);
			current->exit_code = 0;
		}
		if (ptrace_id == PN_STOP_ENTRY && SYSCALL_RETVAL(regs) == -ENOSYS) {
			SYSCALL_RETVAL(regs) = SYSCALL_NR(regs);
			SYSCALL_PC(regs) -= 2;
		} else if (syscall_is(current, regs, rt_sigtimedwait)) {
			if (SYSCALL_RETVAL(regs) == -EAGAIN || SYSCALL_RETVAL(regs) == -EINTR) {
				SYSCALL_RETVAL(regs) = SYSCALL_NR(regs);
				SYSCALL_PC(regs) -= 2;
			}
		}
		goto out_nolock;
	case PN_STOP_FORK:
		/* fork */
		SYSCALL_RETVAL(regs) = current->ptrace_message;
		dprintk("ptrace fork returns pid %ld\n", SYSCALL_RETVAL(regs));
		goto out;
	case PN_STOP_VFORK:
		/* after vfork */
		SYSCALL_RETVAL(regs) = current->ptrace_message;
		dprintk("ptrace after vfork returns pid %ld\n", SYSCALL_RETVAL(regs));
		goto out;
	case PN_STOP_SIGNAL:
		/* normal case : dequeue signal */
		break;
	case PN_STOP_EXIT:
		dprintk("ptrace exit caught\n");
		current->ptrace &= ~PT_TRACE_EXIT;
		spin_unlock_irq(&current->sighand->siglock);
		module_put(THIS_MODULE);
		complete_and_exit(NULL, current->ptrace_message);
		BUG();
	case PN_STOP_EXEC:
		eprintk("ptrace after exec caught: must not happen\n");
		BUG();
	default:
		eprintk("ptrace with unknown identity %d\n", ptrace_id);
		BUG();
	}

	signr = current->exit_code;
	if (signr == 0) {
		dprintk("rlsi: canceled signal %d\n", info->si_signo);
		goto out;
	}
	current->exit_code = 0;

	if (signr != info->si_signo) {
		info->si_signo = signr;
		info->si_errno = 0;
		info->si_code = SI_USER;
		info->si_pid = virt_pid(current->parent);
		info->si_uid = current->parent->uid;
	}

	/* If the (new) signal is now blocked, requeue it.  */
	if (sigismember(&current->blocked, signr)) {
		dprintk("going to requeue signal %d\n", signr);
		goto out_resend_sig;
	}

	ka = &current->sighand->action[signr-1];
	if (ka->sa.sa_handler == SIG_IGN) {
		dprintk("going to resend signal %d (ignored)\n", signr);
		goto out;
	}
	if (ka->sa.sa_handler != SIG_DFL) {
		dprintk("going to resend signal %d (not SIG_DFL)\n", signr);
		goto out_resend_sig;
	}
        if (signr == SIGCONT ||
	    signr == SIGCHLD ||
	    signr == SIGWINCH ||
	    signr == SIGURG ||
	    current->pid == 1)
		goto out;

	/* All the rest, which we cannot handle are requeued. */
	dprintk("going to resend signal %d (sigh)\n", signr);
out_resend_sig:
	spin_unlock_irq(&current->sighand->siglock);
	send_sig_info(signr, info, current);
	module_put(THIS_MODULE);
	return (unsigned long)(info+1);

out:
	spin_unlock_irq(&current->sighand->siglock);
out_nolock:
	module_put(THIS_MODULE);
	return (unsigned long)(info+1);
}

static void ret_finish_stop(void)
{
	/* ...
	 * do_signal() ->
	 *   get_signal_to_deliver() ->
	 *     do_signal_stop() ->
	 *       finish_stop()
	 *
	 * Normally after SIGCONT it will dequeue the next signal. If no signal
	 * is found, do_signal restarts syscall unconditionally.
	 * Otherwise signal handler is pushed on user stack.
	 */

	dprintk("rfs: " CPT_FID "\n", CPT_TID(current));

	clear_stop_state(current);
	current->exit_code = 0;

	module_put(THIS_MODULE);
}

static void ret_restart_sys(void)
{
	struct pt_regs *regs = task_pt_regs(current);

	/* This hook is supposed to be executed, when we have
	 * to complete some interrupted syscall.
	 */
	dprintk("rrs: " CPT_FID "\n", CPT_TID(current));

	if (syscall_is(current,regs,pause)) {
		if (SYSCALL_RETVAL(regs) == -ERESTARTNOHAND) {
			current->state = TASK_INTERRUPTIBLE;
			schedule();
		}
	} else if (syscall_is(current,regs,rt_sigtimedwait)) {
		if (SYSCALL_RETVAL(regs) == -EAGAIN || SYSCALL_RETVAL(regs) == -EINTR) {
			SYSCALL_RETVAL(regs) = SYSCALL_NR(regs);
			SYSCALL_PC(regs) -= 2;
		}
	} else if (syscall_is(current,regs,futex)) {
		if (SYSCALL_RETVAL(regs) == -EINTR) {
			SYSCALL_RETVAL(regs) = SYSCALL_NR(regs);
			SYSCALL_PC(regs) -= 2;
		}
	}

	if (!signal_pending(current)) {
		if (SYSCALL_RETVAL(regs) == -ERESTARTSYS ||
		    SYSCALL_RETVAL(regs) == -ERESTARTNOINTR ||
		    SYSCALL_RETVAL(regs) == -ERESTARTNOHAND) {
			SYSCALL_RETVAL(regs) = SYSCALL_NR(regs);
			SYSCALL_PC(regs) -= 2;
		} else if (SYSCALL_RETVAL(regs) == -ERESTART_RESTARTBLOCK) {
			SYSCALL_RETVAL(regs) = __NR_restart_syscall;
#ifdef CONFIG_X86_64
			if (current->thread_info->flags&_TIF_IA32)
				SYSCALL_RETVAL(regs) = __NR32_restart_syscall;
#endif
			SYSCALL_PC(regs) -= 2;
		}
	}

	module_put(THIS_MODULE);
}

extern void ret_last_siginfo(void);
extern void ret_child_tid(void);
extern void ret_from_rst(void);
extern void pre_ret_from_fork(void);

#ifndef CONFIG_X86_64

/* tsk->thread.eip points to pre_ret_from_fork
 * Stack layout:
 * [eip of the last hook]
 * [args of the last hook]
 * [eip of previous hook]
 * [args of previous hook]
 * ...
 * [eip of the first hook]
 * [args of the first hook]
 * [ret_from_rst]
 */

static void * add_hook(task_t *tsk, void (*hook)(void), int argsize, int *hooks)
{
	ESP(tsk) -= sizeof(unsigned long);
	*(unsigned long*)ESP(tsk) = tsk->thread.eip;
	ESP(tsk) -= argsize;
	tsk->thread.eip = (unsigned long)hook;
	if (!try_module_get(THIS_MODULE)) BUG();
	(*hooks)++;
	return (void*)ESP(tsk);
}

static int restore_registers(task_t *tsk, struct pt_regs *regs,
			     struct cpt_task_image *ti, struct cpt_x86_regs *b)
{
	if (b->cpt_object != CPT_OBJ_X86_REGS)
		return -EINVAL;

	tsk->thread.esp = (unsigned long) regs;
	tsk->thread.esp0 = (unsigned long) (regs+1);
	tsk->thread.eip = (unsigned long) ret_from_rst;

	tsk->thread.fs = decode_segment(b->cpt_fs);
	tsk->thread.gs = decode_segment(b->cpt_gs);
	tsk->thread.debugreg[0] = b->cpt_debugreg[0];
	tsk->thread.debugreg[1] = b->cpt_debugreg[1];
	tsk->thread.debugreg[2] = b->cpt_debugreg[2];
	tsk->thread.debugreg[3] = b->cpt_debugreg[3];
	tsk->thread.debugreg[4] = b->cpt_debugreg[4];
	tsk->thread.debugreg[5] = b->cpt_debugreg[5];
	tsk->thread.debugreg[6] = b->cpt_debugreg[6];
	tsk->thread.debugreg[7] = b->cpt_debugreg[7];

	memcpy(regs, &b->cpt_ebx, sizeof(struct pt_regs));

	regs->xcs = decode_segment(b->cpt_xcs);
	regs->xss = decode_segment(b->cpt_xss);
	regs->xds = decode_segment(b->cpt_xds);
	regs->xes = decode_segment(b->cpt_xes);

	return 0;
}

#else

/* Stack layout:
 *
 * [eip of the last hook]
 * [args of the last hook]
 * ...
 * [eip of the first hook]
 * [args of the first hook]
 * [ret_from_fork+5]
 */

static void * add_hook(task_t *tsk, void (*hook)(void), int argsize, int *hooks)
{
	if (!*hooks) {
		extern void ret_from_fork2(void);
		ESP(tsk) -= sizeof(unsigned long);
		*(unsigned long*)ESP(tsk) = (unsigned long)ret_from_fork2;
		tsk->thread_info->flags |= _TIF_RESUME;
	}
	ESP(tsk) -= argsize + sizeof(unsigned long);
	*(unsigned long*)ESP(tsk) = (unsigned long)hook;
	if (!try_module_get(THIS_MODULE)) BUG();
	(*hooks)++;
	return (void*)(ESP(tsk) + sizeof(unsigned long));
}

static void xlate_ptregs_32_to_64(struct pt_regs *d, struct cpt_x86_regs *s)
{
	memset(d, 0, sizeof(struct pt_regs));
	d->rbp = s->cpt_ebp;
	d->rbx = s->cpt_ebx;
	d->rax = (s32)s->cpt_eax;
	d->rcx = s->cpt_ecx;
	d->rdx = s->cpt_edx;
	d->rsi = s->cpt_esi;
	d->rdi = s->cpt_edi;
	d->orig_rax = (s32)s->cpt_orig_eax;
	d->rip = s->cpt_eip;
	d->cs = s->cpt_xcs;
	d->eflags = s->cpt_eflags;
	d->rsp = s->cpt_esp;
	d->ss = s->cpt_xss;
}

static int restore_registers(task_t *tsk, struct pt_regs *regs,
			     struct cpt_task_image *ti, struct cpt_obj_bits *hdr)
{
	if (hdr->cpt_object == CPT_OBJ_X86_64_REGS) {
		struct cpt_x86_64_regs *b = (void*)hdr;

		tsk->thread.rsp = (unsigned long) regs;
		tsk->thread.rsp0 = (unsigned long) (regs+1);

		tsk->thread.fs = b->cpt_fsbase;
		tsk->thread.gs = b->cpt_gsbase;
		tsk->thread.fsindex = decode_segment(b->cpt_fsindex);
		tsk->thread.gsindex = decode_segment(b->cpt_gsindex);
		tsk->thread.ds = decode_segment(b->cpt_ds);
		tsk->thread.es = decode_segment(b->cpt_es);
		tsk->thread.debugreg0 = b->cpt_debugreg[0];
		tsk->thread.debugreg1 = b->cpt_debugreg[1];
		tsk->thread.debugreg2 = b->cpt_debugreg[2];
		tsk->thread.debugreg3 = b->cpt_debugreg[3];
		tsk->thread.debugreg6 = b->cpt_debugreg[6];
		tsk->thread.debugreg7 = b->cpt_debugreg[7];

		memcpy(regs, &b->cpt_r15, sizeof(struct pt_regs));

		tsk->thread.userrsp = regs->rsp;
		regs->cs = decode_segment(b->cpt_cs);
		regs->ss = decode_segment(b->cpt_ss);
	} else if (hdr->cpt_object == CPT_OBJ_X86_REGS) {
		struct cpt_x86_regs *b = (void*)hdr;

		tsk->thread.rsp = (unsigned long) regs;
		tsk->thread.rsp0 = (unsigned long) (regs+1);

		tsk->thread.fs = 0;
		tsk->thread.gs = 0;
		tsk->thread.fsindex = decode_segment(b->cpt_fs);
		tsk->thread.gsindex = decode_segment(b->cpt_gs);
		tsk->thread.debugreg0 = b->cpt_debugreg[0];
		tsk->thread.debugreg1 = b->cpt_debugreg[1];
		tsk->thread.debugreg2 = b->cpt_debugreg[2];
		tsk->thread.debugreg3 = b->cpt_debugreg[3];
		tsk->thread.debugreg6 = b->cpt_debugreg[6];
		tsk->thread.debugreg7 = b->cpt_debugreg[7];

		xlate_ptregs_32_to_64(regs, b);

		tsk->thread.userrsp = regs->rsp;
		regs->cs = decode_segment(b->cpt_xcs);
		regs->ss = decode_segment(b->cpt_xss);
		tsk->thread.ds = decode_segment(b->cpt_xds);
		tsk->thread.es = decode_segment(b->cpt_xes);
	} else {
		return -EINVAL;
	}
	return 0;
}

#endif

int rst_restore_process(struct cpt_context *ctx)
{
	cpt_object_t *obj;

	for_each_object(obj, CPT_OBJ_TASK) {
		task_t *tsk = obj->o_obj;
		struct cpt_task_image *ti = obj->o_image;
		struct pt_regs * regs;
		struct cpt_object_hdr *b;
		struct cpt_siginfo_image *lsi = NULL;
		struct group_info *gids, *ogids;
		int hooks = 0;
		int i;

		if (tsk == NULL) {
			eprintk_ctx("oops, task %d/%s is missing\n", ti->cpt_pid, ti->cpt_comm);
			return -EFAULT;
		}

		wait_task_inactive(tsk);
		regs = task_pt_regs(tsk);

		if (!tsk->exit_state) {
			tsk->lock_depth = -1;
#ifdef CONFIG_PREEMPT
			tsk->thread_info->preempt_count--;
#endif
		}

		if (tsk->static_prio != ti->cpt_static_prio)
			set_user_nice(tsk, PRIO_TO_NICE(ti->cpt_static_prio));

		cpt_sigset_import(&tsk->blocked, ti->cpt_sigblocked);
		cpt_sigset_import(&tsk->real_blocked, ti->cpt_sigrblocked);
		cpt_sigset_import(&tsk->saved_sigmask, ti->cpt_sigsuspend_blocked);
		cpt_sigset_import(&tsk->pending.signal, ti->cpt_sigpending);

		tsk->uid = ti->cpt_uid;
		tsk->euid = ti->cpt_euid;
		tsk->suid = ti->cpt_suid;
		tsk->fsuid = ti->cpt_fsuid;
		tsk->gid = ti->cpt_gid;
		tsk->egid = ti->cpt_egid;
		tsk->sgid = ti->cpt_sgid;
		tsk->fsgid = ti->cpt_fsgid;
		memcpy(&tsk->cap_effective, &ti->cpt_ecap, sizeof(tsk->cap_effective));
		memcpy(&tsk->cap_inheritable, &ti->cpt_icap, sizeof(tsk->cap_inheritable));
		memcpy(&tsk->cap_permitted, &ti->cpt_pcap, sizeof(tsk->cap_permitted));
		tsk->keep_capabilities = (ti->cpt_keepcap != 0);
		tsk->did_exec = (ti->cpt_did_exec != 0);
		gids = groups_alloc(ti->cpt_ngids);
		ogids = tsk->group_info;
		if (gids) {
			int i;
			for (i=0; i<32; i++)
				gids->small_block[i] = ti->cpt_gids[i];
			tsk->group_info = gids;
		}
		if (ogids)
			put_group_info(ogids);
		tsk->utime = ti->cpt_utime;
		tsk->stime = ti->cpt_stime;
		if (ctx->image_version == 0) {
			tsk->start_time = _ns_to_timespec(ti->cpt_starttime*TICK_NSEC);
		} else {
			cpt_timespec_import(&tsk->start_time, ti->cpt_starttime);
		}
		_set_normalized_timespec(&tsk->start_time,
					tsk->start_time.tv_sec -
					get_exec_env()->init_entry->start_time.tv_sec,
					tsk->start_time.tv_nsec -
					get_exec_env()->init_entry->start_time.tv_nsec);

		tsk->nvcsw = ti->cpt_nvcsw;
		tsk->nivcsw = ti->cpt_nivcsw;
		tsk->min_flt = ti->cpt_min_flt;
		tsk->maj_flt = ti->cpt_maj_flt;

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,6,8)
		tsk->cutime = ti->cpt_cutime;
		tsk->cstime = ti->cpt_cstime;
		tsk->cnvcsw = ti->cpt_cnvcsw;
		tsk->cnivcsw = ti->cpt_cnivcsw;
		tsk->cmin_flt = ti->cpt_cmin_flt;
		tsk->cmaj_flt = ti->cpt_cmaj_flt;

		if (RLIM_NLIMITS > CPT_RLIM_NLIMITS)
			__asm__("undefined\n");

		for (i=0; i<RLIM_NLIMITS; i++) {
			tsk->rlim[i].rlim_cur = ti->cpt_rlim_cur[i];
			tsk->rlim[i].rlim_max = ti->cpt_rlim_max[i];
		}
#else
		if (thread_group_leader(tsk) && tsk->signal) {
			tsk->signal->utime = ti->cpt_utime;
			tsk->signal->stime = ti->cpt_stime;
			tsk->signal->cutime = ti->cpt_cutime;
			tsk->signal->cstime = ti->cpt_cstime;
			tsk->signal->nvcsw = ti->cpt_nvcsw;
			tsk->signal->nivcsw = ti->cpt_nivcsw;
			tsk->signal->cnvcsw = ti->cpt_cnvcsw;
			tsk->signal->cnivcsw = ti->cpt_cnivcsw;
			tsk->signal->min_flt = ti->cpt_min_flt;
			tsk->signal->maj_flt = ti->cpt_maj_flt;
			tsk->signal->cmin_flt = ti->cpt_cmin_flt;
			tsk->signal->cmaj_flt = ti->cpt_cmaj_flt;

			if (RLIM_NLIMITS > CPT_RLIM_NLIMITS)
				__asm__("undefined\n");

			for (i=0; i<RLIM_NLIMITS; i++) {
				tsk->signal->rlim[i].rlim_cur = ti->cpt_rlim_cur[i];
				tsk->signal->rlim[i].rlim_max = ti->cpt_rlim_max[i];
			}
		}
#endif

		for (i=0; i<3; i++) {
			if (i >= GDT_ENTRY_TLS_ENTRIES) {
				eprintk_ctx("too many tls descs\n");
			} else {
#ifndef CONFIG_X86_64
				tsk->thread.tls_array[i].a = ti->cpt_tls[i]&0xFFFFFFFF;
				tsk->thread.tls_array[i].b = ti->cpt_tls[i]>>32;
#else
				tsk->thread.tls_array[i] = ti->cpt_tls[i];
#endif
			}
		}

		clear_stopped_child_used_math(tsk);

		b = (void *)(ti+1);
		while ((void*)b < ((void*)ti) + ti->cpt_next) {
			/* Siginfo objects are at the end of obj array */
			if (b->cpt_object == CPT_OBJ_SIGINFO) {
				struct ve_struct *env = set_exec_env(VE_TASK_INFO(tsk)->owner_env);
				restore_sigqueue(tsk, &tsk->pending, (unsigned long)b, (unsigned long)ti + ti->cpt_next);
				set_exec_env(env);
				break;
			}

			switch (b->cpt_object) {
			case CPT_OBJ_BITS:
				if (b->cpt_content == CPT_CONTENT_X86_FPUSTATE &&
				    cpu_has_fxsr) {
					memcpy(&tsk->thread.i387,
					       (void*)b + b->cpt_hdrlen,
					       sizeof(struct i387_fxsave_struct));
					if (ti->cpt_used_math)
						set_stopped_child_used_math(tsk);
				}
#ifdef CONFIG_X86_32
				else if (b->cpt_content == CPT_CONTENT_X86_FPUSTATE_OLD &&
					 !cpu_has_fxsr) {
					memcpy(&tsk->thread.i387,
					       (void*)b + b->cpt_hdrlen,
					       sizeof(struct i387_fsave_struct));
					if (ti->cpt_used_math)
						set_stopped_child_used_math(tsk);
				}
#endif
				break;
			case CPT_OBJ_LASTSIGINFO:
				lsi = (void*)b;
				break;
			case CPT_OBJ_X86_REGS:
			case CPT_OBJ_X86_64_REGS:
				if (restore_registers(tsk, regs, ti, (void*)b)) {
					eprintk_ctx("cannot restore registers: image is corrupted\n");
					return -EINVAL;
				}
				break;
			case CPT_OBJ_SIGALTSTACK: {
				struct cpt_sigaltstack_image *sas;
				sas = (struct cpt_sigaltstack_image *)b;
				tsk->sas_ss_sp = sas->cpt_stack;
				tsk->sas_ss_size = sas->cpt_stacksize;
				break;
			    }
			}
			b = ((void*)b) + b->cpt_next;
		}

		if (ti->cpt_ppid != ti->cpt_rppid) {
			task_t *parent;
			struct ve_struct *env = set_exec_env(VE_TASK_INFO(tsk)->owner_env);
			write_lock_irq(&tasklist_lock);
			parent = find_task_by_pid_ve(ti->cpt_ppid);
			if (parent && parent != tsk->parent) {
				list_add(&tsk->ptrace_list, &tsk->parent->ptrace_children);
				REMOVE_LINKS(tsk);
				tsk->parent = parent;
				SET_LINKS(tsk);
			}
			write_unlock_irq(&tasklist_lock);
			set_exec_env(env);
		}

		tsk->ptrace_message = ti->cpt_ptrace_message;
		tsk->pn_state = ti->cpt_pn_state;
		tsk->stopped_state = ti->cpt_stopped_state;
		tsk->thread_info->flags = ti->cpt_thrflags;

		/* The image was created with kernel < 2.6.16, while
		 * task hanged in sigsuspend -> do_signal.
		 *
		 * FIXME! This needs more brain efforts...
		 */
		if (ti->cpt_sigsuspend_state) {
			tsk->thread_info->flags |= _TIF_RESTORE_SIGMASK;
		}

#ifdef CONFIG_X86_64
		tsk->thread_info->flags |= _TIF_FORK;
		if (!ti->cpt_64bit)
			tsk->thread_info->flags |= _TIF_IA32;
#endif

#ifndef CONFIG_X86_64
		do {
			if (regs->orig_eax == __NR__newselect && regs->edi) {
				struct timeval tv;
				if (access_process_vm(tsk, regs->edi, &tv,
						sizeof(tv), 0) != sizeof(tv)) {
					wprintk_ctx("task %d/%d(%s): Error 1 in access_process_vm: edi %ld\n",
						virt_pid(tsk), tsk->pid, tsk->comm,
					       regs->edi);
					break;
				}
				dprintk_ctx("task %d/%d(%s): Old timeval in newselect: %ld.%ld\n",
				       virt_pid(tsk), tsk->pid, tsk->comm,
				       tv.tv_sec, tv.tv_usec);
				tv.tv_sec -= ctx->delta_time.tv_sec;
				if (tv.tv_usec < ctx->delta_time.tv_nsec / 1000) {
					tv.tv_usec += 1000000 - ctx->delta_time.tv_nsec / 1000;
					tv.tv_sec--;
				} else {
					tv.tv_usec -= ctx->delta_time.tv_nsec / 1000;
				}
				if (tv.tv_sec < 0) {
					tv.tv_sec = 0;
					tv.tv_usec = 0;
				}
				dprintk_ctx("task %d/%d(%s): New timeval in newselect: %ld.%ld\n",
					virt_pid(tsk), tsk->pid, tsk->comm,
				       tv.tv_sec, tv.tv_usec);
				if (access_process_vm(tsk, regs->edi, &tv,
						sizeof(tv), 1) != sizeof(tv)) {
					wprintk_ctx("task %d/%d(%s): Error 1 in access_process_vm write: edi %ld\n",
						virt_pid(tsk), tsk->pid, tsk->comm, regs->edi);
				}

			} else if (regs->orig_eax == __NR_select && regs->edi) {
				struct {
					unsigned long n;
					fd_set __user *inp, *outp, *exp;
					struct timeval __user *tvp;
				} a;
				struct timeval tv;
				if (access_process_vm(tsk, regs->ebx, &a,
						sizeof(a), 0) != sizeof(a)) {
					wprintk_ctx("task %d: Error 2 in access_process_vm\n", tsk->pid);
					break;
				}
				if (access_process_vm(tsk, (unsigned long)a.tvp,
						&tv, sizeof(tv), 0) != sizeof(tv)) {
					wprintk_ctx("task %d: Error 3 in access_process_vm\n", tsk->pid);
					break;
				}
				dprintk_ctx("task %d: Old timeval in select: %ld.%ld\n",
					tsk->pid, tv.tv_sec, tv.tv_usec);
				tv.tv_sec -= ctx->delta_time.tv_sec;
				if (tv.tv_usec < ctx->delta_time.tv_nsec / 1000) {
					tv.tv_usec += 1000000 - ctx->delta_time.tv_nsec / 1000;
					tv.tv_sec--;
				} else {
					tv.tv_usec -= ctx->delta_time.tv_nsec / 1000;
				}
				if (tv.tv_sec < 0) {
					tv.tv_sec = 0;
					tv.tv_usec = 0;
				}
				dprintk_ctx("task %d: New timeval in select: %ld.%ld\n",
					tsk->pid, tv.tv_sec, tv.tv_usec);
				if (access_process_vm(tsk, (unsigned long)a.tvp,
						&tv, sizeof(tv), 1) != sizeof(tv)) {
					wprintk_ctx("task %d: Error 3 in access_process_vm write\n", tsk->pid);
				}
			}
		} while (0);
#endif

		if (!tsk->exit_state && (long)SYSCALL_NR(regs) >= 0) {
			if (SYSCALL_RETVAL(regs) == -ERESTARTSYS ||
			    SYSCALL_RETVAL(regs) == -ERESTARTNOINTR ||
			    SYSCALL_RETVAL(regs) == -ERESTARTNOHAND ||
			    SYSCALL_RETVAL(regs) == -ERESTART_RESTARTBLOCK ||
			    syscall_is(tsk,regs,pause) ||
			    (syscall_is(tsk,regs,rt_sigtimedwait) &&
			     (SYSCALL_RETVAL(regs) == -EAGAIN || SYSCALL_RETVAL(regs) == -EINTR)) ||
			    (syscall_is(tsk,regs,futex) &&
			     (SYSCALL_RETVAL(regs) == -EINTR)))
				add_hook(tsk, ret_restart_sys, 0, &hooks);
		}

		if (lsi || tsk->pn_state) {
			/* ... -> ptrace_notify()
			 * or
			 * ... -> do_signal() -> get_signal_to_deliver() ->
			 *   ptrace stop
			 */
			tsk->last_siginfo = add_hook(tsk, ret_last_siginfo, sizeof(siginfo_t), &hooks);
			memset(tsk->last_siginfo, 0, sizeof(siginfo_t));
			if (lsi)
				decode_siginfo(tsk->last_siginfo, lsi);
		}

		tsk->ptrace = ti->cpt_ptrace;
		tsk->flags = ti->cpt_flags & ~PF_FROZEN;
		clear_tsk_thread_flag(tsk, TIF_FREEZE);
		tsk->exit_signal = ti->cpt_exit_signal;

		if (tsk->stopped_state) {
			dprintk_ctx("finish_stop\n");
			if (ti->cpt_state != TASK_STOPPED)
				eprintk_ctx("Hellooo, state is %u\n", (unsigned)ti->cpt_state);
			add_hook(tsk, ret_finish_stop, 0, &hooks);
		}

		if (!tsk->exit_state &&
		    (ti->cpt_set_tid || ti->cpt_clear_tid)) {
			unsigned long *ptr = add_hook(tsk, ret_child_tid, sizeof(unsigned long)*2, &hooks);
			ptr[0] = ti->cpt_clear_tid;
			ptr[1] = ti->cpt_set_tid;
			dprintk_ctx("settids\n");
		}

#ifdef CONFIG_X86_64
		if (!hooks && (long)SYSCALL_NR(regs) < 0) {
			extern void ret_from_fork2(void);
			ESP(tsk) -= sizeof(unsigned long);
			*(unsigned long*)ESP(tsk) = (unsigned long)ret_from_fork2;
			tsk->thread_info->flags |= _TIF_RESUME;
		}
#else
		tsk->thread.esp -= 4;
		*(__u32*)tsk->thread.esp = tsk->thread.eip;
		tsk->thread.eip = (unsigned long)pre_ret_from_fork;
#endif

		if (ti->cpt_state == TASK_TRACED)
			tsk->state = TASK_TRACED;
		else if (ti->cpt_state & (EXIT_ZOMBIE|EXIT_DEAD)) {
			tsk->signal->it_virt_expires = 0;
			tsk->signal->it_prof_expires = 0;
			if (tsk->state != EXIT_DEAD)
				eprintk_ctx("oops, schedule() did not make us dead\n");
		}

		if (thread_group_leader(tsk) &&
		    ti->cpt_it_real_value &&
		    !(ti->cpt_state & (EXIT_ZOMBIE|EXIT_DEAD))) {
			DEFINE_KTIME(val);

			if (ctx->image_version != 0) {
				ktime_t delta;

				val = ktime_add_ns(val, ti->cpt_it_real_value);
				delta = timespec_to_ktime(ctx->delta_time);
				val = ktime_sub(val, delta);
				if (val.tv64 <= 0)
					val.tv64 = NSEC_PER_USEC;
				dprintk("rst itimer " CPT_FID " +%Ld %Ld %Lu\n", CPT_TID(tsk), val.tv64, delta.tv64, ti->cpt_it_real_value);
			} else {
				unsigned long jif = ti->cpt_it_real_value -
					timespec_to_jiffies(&ctx->delta_time);
				if ((long)jif <= 0)
					jif = 1;
				val = ktime_add_ns(val, (u64)jif*TICK_NSEC);
			}
			spin_lock_irq(&tsk->sighand->siglock);
			if (hrtimer_try_to_cancel(&tsk->signal->real_timer) >= 0) {
				/* FIXME. Check!!!! */
				hrtimer_start(&tsk->signal->real_timer, val, HRTIMER_REL);
			} else {
				wprintk_ctx("Timer clash. Impossible?\n");
			}
			spin_unlock_irq(&tsk->sighand->siglock);

			dprintk_ctx("itimer " CPT_FID " +%Lu\n", CPT_TID(tsk), val.tv64);
		}

		module_put(THIS_MODULE);
	}
	return 0;
}
