/*
 *
 *  kernel/cpt/rst_sysvipc.c
 *
 *  Copyright (C) 2000-2005  SWsoft
 *  All rights reserved.
 *
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/mm.h>
#include <linux/errno.h>
#include <linux/major.h>
#include <linux/pipe_fs_i.h>
#include <linux/mman.h>
#include <linux/shm.h>
/* FIXME. x86_64 has asm/ipc.h forgotten? */
#include <asm-generic/ipc.h>
#include <asm/uaccess.h>
#include <asm/unistd.h>
#include <ub/ub_mem.h>
#include <linux/cpt_image.h>

#include "cpt_obj.h"
#include "cpt_context.h"
#include "cpt_kernel.h"

struct _warg {
		struct file		*file;
		struct cpt_sysvshm_image	*v;
};

static int fixup_one_shm(struct shmid_kernel *shp, void *arg)
{
	struct _warg *warg = arg;

	if (shp->shm_file != warg->file)
		return 0;
	if (shp->shm_nattch)
		return -EEXIST;

	shp->shm_perm.uid = warg->v->cpt_uid;
	shp->shm_perm.gid = warg->v->cpt_gid;
	shp->shm_perm.cuid = warg->v->cpt_cuid;
	shp->shm_perm.cgid = warg->v->cpt_cgid;
	shp->shm_perm.mode = warg->v->cpt_mode;

	shp->shm_atim = warg->v->cpt_atime;
	shp->shm_dtim = warg->v->cpt_dtime;
	shp->shm_ctim = warg->v->cpt_ctime;
	shp->shm_cprid = warg->v->cpt_creator;
	shp->shm_lprid = warg->v->cpt_last;

	/* TODO: fix shp->mlock_user? */
	return 1;
}

static int fixup_shm(struct file *file, struct cpt_sysvshm_image *v)
{
	struct _warg warg;

	warg.file = file;
	warg.v = v;

	return sysvipc_walk_shm(fixup_one_shm, &warg);
}

static int fixup_shm_data(struct file *file, loff_t pos, loff_t end,
			  struct cpt_context *ctx)
{
	struct cpt_page_block pgb;
	ssize_t (*do_write)(struct file *, const char __user *, size_t, loff_t *ppos);

	do_write = file->f_dentry->d_inode->i_fop->write;
	if (do_write == NULL) {
		eprintk_ctx("No TMPFS? Cannot restore content of SYSV SHM\n");
		return -EINVAL;
	}

	while (pos < end) {
		loff_t opos;
		loff_t ipos;
		int count;
		int err;

		err = rst_get_object(CPT_OBJ_PAGES, pos, &pgb, ctx);
		if (err)
			return err;
		dprintk_ctx("restoring SHM block: %08x-%08x\n",
		       (__u32)pgb.cpt_start, (__u32)pgb.cpt_end);
		ipos = pos + pgb.cpt_hdrlen;
		opos = pgb.cpt_start;
		count = pgb.cpt_end-pgb.cpt_start;
		while (count > 0) {
			mm_segment_t oldfs;
			int copy = count;

			if (copy > PAGE_SIZE)
				copy = PAGE_SIZE;
			(void)cpt_get_buf(ctx);
			oldfs = get_fs(); set_fs(KERNEL_DS);
			err = ctx->pread(ctx->tmpbuf, copy, ctx, ipos);
			set_fs(oldfs);
			if (err) {
				__cpt_release_buf(ctx);
				return err;
			}
			oldfs = get_fs(); set_fs(KERNEL_DS);
			ipos += copy;
			err = do_write(file, ctx->tmpbuf, copy, &opos);
			set_fs(oldfs);
			__cpt_release_buf(ctx);
			if (err != copy) {
				eprintk_ctx("write() failure\n");
				if (err >= 0)
					err = -EIO;
				return err;
			}
			count -= copy;
		}
		pos += pgb.cpt_next;
	}
	return 0;
}

struct file * rst_sysv_shm(loff_t pos, struct cpt_context *ctx)
{
	struct file *file;
	int err;
	loff_t dpos, epos;
	union {
		struct cpt_file_image		fi;
		struct cpt_sysvshm_image	shmi;
		struct cpt_inode_image 		ii;
	} u;

	err = rst_get_object(CPT_OBJ_FILE, pos, &u.fi, ctx);
	if (err < 0)
		goto err_out;
	pos = u.fi.cpt_inode;
	err = rst_get_object(CPT_OBJ_INODE, pos, &u.ii, ctx);
	if (err < 0)
		goto err_out;
	dpos = pos + u.ii.cpt_hdrlen;
	epos = pos + u.ii.cpt_next;
	err = rst_get_object(CPT_OBJ_SYSV_SHM, pos + u.ii.cpt_hdrlen, &u.shmi, ctx);
	if (err < 0)
		goto err_out;
	dpos += u.shmi.cpt_next;

	file = sysvipc_setup_shm(u.shmi.cpt_key, u.shmi.cpt_id,
				 u.shmi.cpt_segsz, u.shmi.cpt_mode);
	if (!IS_ERR(file)) {
		err = fixup_shm(file, &u.shmi);
		if (err != -EEXIST && dpos < epos)
			err = fixup_shm_data(file, dpos, epos, ctx);
	}

	return file;

err_out:
	return ERR_PTR(err);
}

static int attach_one_undo(int semid, struct sem_array *sma, void *arg)
{
	struct sem_undo *su = arg;
	struct sem_undo_list *undo_list = current->sysvsem.undo_list;

	if (semid != su->semid)
		return 0;

	su->proc_next = undo_list->proc_list;
	undo_list->proc_list = su;

	su->id_next = sma->undo;
	sma->undo = su;

	return 1;
}

static int attach_undo(struct sem_undo *su)
{
	return sysvipc_walk_sem(attach_one_undo, su);
}

static int do_rst_semundo(struct cpt_object_hdr *sui, loff_t pos, struct cpt_context *ctx)
{
	int err;
	struct sem_undo_list *undo_list;

	if (current->sysvsem.undo_list) {
		eprintk_ctx("Funny undo_list\n");
		return 0;
	}

	undo_list = ub_kmalloc(sizeof(struct sem_undo_list), GFP_KERNEL);
	if (undo_list == NULL)
		return -ENOMEM;
	memset(undo_list, 0, sizeof(struct sem_undo_list));
	atomic_set(&undo_list->refcnt, 1);
	spin_lock_init(&undo_list->lock);
	current->sysvsem.undo_list = undo_list;

	if (sui->cpt_next > sui->cpt_hdrlen) {
		loff_t offset = pos + sui->cpt_hdrlen;
		do {
			struct sem_undo *new;
			struct cpt_sysvsem_undo_image spi;
			err = rst_get_object(CPT_OBJ_SYSVSEM_UNDO_REC, offset, &spi, ctx);
			if (err)
				goto out;
			new = ub_kmalloc(sizeof(struct sem_undo) +
					 sizeof(short)*spi.cpt_nsem, GFP_KERNEL);
			if (!new) {
				err = -ENOMEM;
				goto out;
			}

			memset(new, 0, sizeof(struct sem_undo) + sizeof(short)*spi.cpt_nsem);
			new->semadj = (short *) &new[1];
			new->semid = spi.cpt_id;
			err = ctx->pread(new->semadj, spi.cpt_nsem*sizeof(short), ctx, offset + spi.cpt_hdrlen);
			if (err) {
				kfree(new);
				goto out;
			}
			err = attach_undo(new);
			if (err <= 0) {
				if (err == 0)
					err = -ENOENT;
				kfree(new);
				goto out;
			}
			offset += spi.cpt_next;
		} while (offset < pos + sui->cpt_next);
	}
	err = 0;

out:
	return err;
}

__u32 rst_semundo_flag(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	__u32 flag = 0;

#if 0
	if (ti->cpt_sysvsem_undo == CPT_NULL ||
	    lookup_cpt_obj_bypos(CPT_OBJ_SYSVSEM_UNDO, ti->cpt_sysvsem_undo))
		flag |= CLONE_SYSVSEM;
#endif
	return flag;
}

int rst_semundo_complete(struct cpt_task_image *ti, struct cpt_context *ctx)
{
	int err;
	struct sem_undo_list *f = current->sysvsem.undo_list;
	cpt_object_t *obj;
	struct cpt_object_hdr sui;

	if (ti->cpt_sysvsem_undo == CPT_NULL) {
		exit_sem(current);
		return 0;
	}

	obj = lookup_cpt_obj_bypos(CPT_OBJ_SYSVSEM_UNDO, ti->cpt_sysvsem_undo, ctx);
	if (obj) {
		if (obj->o_obj != f) {
			exit_sem(current);
			f = obj->o_obj;
			atomic_inc(&f->refcnt);
			current->sysvsem.undo_list = f;
		}
		return 0;
	}

	if ((err = rst_get_object(CPT_OBJ_SYSVSEM_UNDO, ti->cpt_sysvsem_undo, &sui, ctx)) != 0)
		goto out;

	if ((err = do_rst_semundo(&sui, ti->cpt_sysvsem_undo, ctx)) != 0)
		goto out;

	err = -ENOMEM;
	obj = cpt_object_add(CPT_OBJ_SYSVSEM_UNDO, f, ctx);
	if (obj) {
		err = 0;
		cpt_obj_setpos(obj, ti->cpt_sysvsem_undo, ctx);
	}

	return 0;

out:
	return err;
}

struct _sarg {
	int semid;
	struct cpt_sysvsem_image	*v;
	__u32				*arr;
};

static int fixup_one_sem(int semid, struct sem_array *sma, void *arg)
{
	struct _sarg *warg = arg;

	if (semid != warg->semid)
		return 0;

	sma->sem_perm.uid = warg->v->cpt_uid;
	sma->sem_perm.gid = warg->v->cpt_gid;
	sma->sem_perm.cuid = warg->v->cpt_cuid;
	sma->sem_perm.cgid = warg->v->cpt_cgid;
	sma->sem_perm.mode = warg->v->cpt_mode;
	sma->sem_perm.seq = warg->v->cpt_seq;

	sma->sem_ctime = warg->v->cpt_ctime;
	sma->sem_otime = warg->v->cpt_otime;
	memcpy(sma->sem_base, warg->arr, sma->sem_nsems*8);
	return 1;
}

static int fixup_sem(int semid, struct cpt_sysvsem_image *v, __u32 *arr)
{
	struct _sarg warg;

	warg.semid = semid;
	warg.v = v;
	warg.arr = arr;

	return sysvipc_walk_sem(fixup_one_sem, &warg);
}


static int restore_sem(loff_t pos, struct cpt_sysvsem_image *si,
		       struct cpt_context *ctx)
{
	int err;
	__u32 *arr;
	int nsems = (si->cpt_next - si->cpt_hdrlen)/8;

	arr = kmalloc(nsems*8, GFP_KERNEL);
	if (!arr)
		return -ENOMEM;

	err = ctx->pread(arr, nsems*8, ctx, pos+si->cpt_hdrlen);
	if (err)
		goto out;
	err = sysvipc_setup_sem(si->cpt_key, si->cpt_id, nsems, si->cpt_mode);
	if (err < 0) {
		eprintk_ctx("SEM 3\n");
		goto out;
	}
	err = fixup_sem(si->cpt_id, si, arr);
	if (err == 0)
		err = -ESRCH;
	if (err > 0)
		err = 0;
out:
	kfree(arr);
	return err;
}

static int rst_sysv_sem(struct cpt_context *ctx)
{
	int err;
	loff_t sec = ctx->sections[CPT_SECT_SYSV_SEM];
	loff_t endsec;
	struct cpt_section_hdr h;
	struct cpt_sysvsem_image sbuf;

	if (sec == CPT_NULL)
		return 0;

	err = ctx->pread(&h, sizeof(h), ctx, sec);
	if (err)
		return err;
	if (h.cpt_section != CPT_SECT_SYSV_SEM || h.cpt_hdrlen < sizeof(h))
		return -EINVAL;

	endsec = sec + h.cpt_next;
	sec += h.cpt_hdrlen;
	while (sec < endsec) {
		int err;
		err = rst_get_object(CPT_OBJ_SYSV_SEM, sec, &sbuf, ctx);
		if (err)
			return err;
		err = restore_sem(sec, &sbuf, ctx);
		if (err)
			return err;
		sec += sbuf.cpt_next;
	}
	return 0;
}

int rst_sysv_ipc(struct cpt_context *ctx)
{
	return rst_sysv_sem(ctx);
}
